"""Access to native code functions

At some point we'll implement a loader here which can be used to load
objects into threads and return the functions inside them. For now, we
just have skeleton classes that hold pre-loaded functions.

"""
from __future__ import annotations
from rsyscall._raw import ffi # type: ignore
import contextlib
from dataclasses import dataclass
import typing as t
from rsyscall.handle import (
    Borrowable, Pointer, WrittenPointer, FileDescriptor, Task,
    Stack,
)
from rsyscall.sys.mman import MemoryMapping
from rsyscall.memory.allocation_interface import AllocationInterface
from rsyscall.memory.transport import MemoryGateway
import rsyscall.near.types as near
import rsyscall.far as far
from rsyscall.struct import Serializer

__all__ = [
    "NativeFunction",
    "Trampoline",
    "NativeLoader",
]

class NativeFunction:
    """A native-code function

    For now, this is just an opaque marker type, but maybe later we'll
    actually be able to manipulate this as an object.

    """
    pass

@dataclass
class Trampoline(Borrowable):
    "A pointer to a native function plus some arguments passed by register"
    function: Pointer[NativeFunction]
    args: t.List[t.Union[FileDescriptor, WrittenPointer[Borrowable], Pointer, int]]

    def __post_init__(self) -> None:
        if len(self.args) > 6:
            raise Exception("only six arguments can be passed via trampoline")

    def borrow_with(self, stack: contextlib.ExitStack, task: Task) -> None:
        "Borrow the function pointer and the arguments"
        stack.enter_context(self.function.borrow(task))
        for arg in self.args:
            if isinstance(arg, int):
                pass
            elif isinstance(arg, WrittenPointer):
                arg.value.borrow_with(stack, task)
            else:
                stack.enter_context(arg.borrow(task))

class TrampolineSerializer(Serializer[Trampoline]):
    def to_bytes(self, val: Trampoline) -> bytes:
        # at most six arguments
        args: t.List[int] = [0]*6
        for i, arg in enumerate(val.args):
            if isinstance(arg, FileDescriptor):
                args[i] = int(arg)
            elif isinstance(arg, Pointer):
                args[i] = int(arg.near)
            else:
                args[i] = int(arg)
        struct = ffi.new('struct rsyscall_trampoline_stack*', {
            'function': ffi.cast('void*', int(val.function.near)),
            'rdi': args[0],
            'rsi': args[1],
            'rdx': args[2],
            'rcx': args[3],
            'r8':  args[4],
            'r9':  args[5],
        })
        return bytes(ffi.buffer(struct))

class StaticAllocation(AllocationInterface):
    def offset(self) -> int:
        return 0

    def size(self) -> int:
        raise Exception

    def split(self, size: int) -> t.Tuple[AllocationInterface, AllocationInterface]:
        raise Exception

    def merge(self, other: AllocationInterface) -> AllocationInterface:
        raise Exception("can't merge")

    def free(self) -> None:
        pass

class NullGateway(MemoryGateway):
    async def read(self, src: Pointer) -> bytes:
        raise Exception("shouldn't try to read")
    async def write(self, dest: Pointer, data: bytes) -> None:
        raise Exception("shouldn't try to write")

class NativeFunctionSerializer(Serializer[NativeFunction]):
    "Serializes NativeFunctions by throw exceptions if you call to_bytes/from_bytes :)"
    pass

@dataclass
class NativeLoader:
    """Loads native code object into memory where they can be called

    At the moment we don't actually load though - we just hardcode a
    set of available symbols that we use.

    """
    server_func: Pointer[NativeFunction]
    persistent_server_func: Pointer[NativeFunction]
    trampoline_func: Pointer[NativeFunction]
    futex_helper_func: Pointer[NativeFunction]

    @staticmethod
    def make_from_symbols(task: Task, symbols: t.Any) -> NativeLoader:
        """Create a NativeLoader by pulling functions from this "symbols" object by attribute

        This symbols object is either rsyscall._raw.lib, or a cffi struct;
        either way, the attributes are named as we expect in this function.

        """
        def to_handle(cffi_ptr) -> Pointer[NativeFunction]:
            pointer_int = int(ffi.cast('ssize_t', cffi_ptr))
            # TODO we're just making up a memory mapping that this pointer is inside;
            # we should figure out the actual mapping, and the size for that matter.
            mapping = MemoryMapping(task, near.MemoryMapping(pointer_int, 0, 1), far.File())
            return Pointer(mapping, NullGateway(), NativeFunctionSerializer(), StaticAllocation(), NativeFunction)
        return NativeLoader(
            server_func=to_handle(symbols.rsyscall_server),
            persistent_server_func=to_handle(symbols.rsyscall_persistent_server),
            trampoline_func=to_handle(symbols.rsyscall_trampoline),
            futex_helper_func=to_handle(symbols.rsyscall_futex_helper),
        )

    def make_trampoline_stack(self, trampoline: Trampoline) -> Stack[Trampoline]:
        "Make a stack that will invoke the passed trampoline as its immediate action"
        return Stack(self.trampoline_func, trampoline, TrampolineSerializer())



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.fcntl import O
from rsyscall.sys.prctl import *
from rsyscall.sys.capability import *
from rsyscall.sched import CLONE

class TestUser(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.local = local_thread
        self.thr = await self.local.clone(CLONE.NEWUSER)

    async def asyncTearDown(self) -> None:
        await self.thr.exit(0)

    async def test_ambient_caps(self) -> None:
        hdr = await self.thr.ptr(CapHeader())
        data_ptr = await self.thr.task.capget(hdr, await self.thr.malloc(CapData))
        data = await data_ptr.read()
        data.inheritable.add(CAP.SYS_ADMIN)
        await self.thr.task.capset(hdr, await data_ptr.write(data))
        await self.thr.task.prctl(PR.CAP_AMBIENT, PR_CAP_AMBIENT.RAISE, CAP.SYS_ADMIN)



"""`#include <linux/...>`

Headers for general Linux functionality, usually non-standard

All these headers are for Linux-specific functionality, although not
all Linux-specific functionality is in these headers.
"""



"""Concurrency based on `shift`/`reset` and object-oriented effect handlers

We implement a portable `shift`/`reset`[0] for async Python, which works when running
under any supported async runner. We also build several concurrent communication
mechanisms with `shift`/`reset`.

Delimited continuations, such as `shift`/`reset`, can be used to transform a
callback-based concurrency system into a more ergonomic direct-style system.[1]
For example, we can go from:

```
def cb(data):
  if data:
    more_work(data)
file.read_cb(cb)
```

to:

```
data = await shift(file.read_cb)
if data:
  more_work(data)
```

We can then construct an ergonomic direct-style concurrency system by layering
delimited continuations on top of a traditional callback-based concurrency
system. This has a number of advantages.

Instead of a coroutine performing a blocking operation by yielding up to a
global event loop, a coroutine performs a blocking operation by calling `shift`
with any object that accepts a callback.  Context switching between running
coroutines happens automatically, in a distributed fashion, as objects receive
events and call callbacks.  Instead of relying implicitly on a global event loop
and global scheduler, a coroutine explicitly selects what object is responsible
for scheduling it, simply by making a call into that object.  That object may
perform its duties by calling into other objects or by performing blocking
system calls itself, and ultimately resumes the coroutine with the result by
calling the callback.

With callback-based concurrency, an object can guarantee that when callbacks are
registered in a certain order, the underlying operations are performed in that
order; and that when events happen in some order, the registered callbacks are
called in that same order.  This preserves information about event ordering and
allows for much simpler bookkeeping of state.  This guarantee is preserved when
using `shift`/`reset`.

For example, suppose some underlying bit of state is either "True" or "False",
and we read that state on each related operation, and maintain a Python boolean
which is supposed to match the underlying state. If callbacks are called in
order, we know that the underlying state is always equal to the state at the
time of the last callback, so we can just set the tracking bool to the bit we
read. Thus, we can safely perform multiple operations at a time, in parallel,
from unrelated coroutines, with no explicit sequencing in user code. Without
this guarantee, we would have to explicitly track the order in which operations
are performed, or, more easily, lock the underlying state so that we only
perform one operation at a time, reducing parallelism.

Our object-oriented style also gives us effect handlers which are not based on a
stack discipline.  We don't need to be concerned about computations being
performed in or out of the scope of handlers.  Effect handlers are simply
regular garbage collected objects: An effect handler stays alive as long as
there is something with a reference to it.

We also get an object-capability-style effect system for free.  Since we don't
rely on an implicit global event loop, there are no implicit global asynchronous
effects.  To allow a function to perform an asynchronous effect, we pass it (as
a normal argument) the object that implements that effect. The type of an
effectful function is simply a regular function type, accepting one or more
objects with which it can perform effects.[2]

[0]: If you aren't familiar with delimited continuations or shift/reset,
read this tutorial:
http://pllab.is.ocha.ac.jp/~asai/cw2011tutorial/main-e.pdf

[1]: Using delimited continuations to pass callbacks is a well known technique;
here's one post about it: 
http://www.gregrosenblatt.com/writing/reinvert-control-delim-cont.html

[2]: Since we can close over effect handlers (they're regular objects), a single
object or function might abstract over multiple effects. Is this a bad thing?
Some say yes, but it seems like a pretty normal form of abstraction to me.

"""
from dneio.core import shift, reset, Continuation, is_running_directly_under_trio
from dneio.concur import RequestQueue, Event, Future, make_n_in_parallel, run_all



"The subset of functionality in unistd.h which relates to IO"
from __future__ import annotations
from rsyscall._raw import lib # type: ignore
import enum
import typing as t

__all__ = [
    "SEEK",
    "IOFileDescriptor",
    "SeekableFileDescriptor",
]

class SEEK(enum.IntEnum):
    "The whence argument to lseek."
    SET = lib.SEEK_SET
    CUR = lib.SEEK_CUR
    END = lib.SEEK_END
    DATA = lib.SEEK_DATA
    HOLE = lib.SEEK_HOLE

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask
from rsyscall.handle.pointer import Pointer, WrittenPointer, ReadablePointer

T = t.TypeVar('T')
class IOFileDescriptor(BaseFileDescriptor):
    async def read(self, buf: Pointer[T]) -> t.Tuple[ReadablePointer[T], Pointer[T]]:
        """read from a file descriptor

        manpage: read(2)
        """
        self._validate()
        buf.check_address_space(self.task)
        ret = await _read(self.task.sysif, self.near, buf.near, buf.size())
        return buf.readable_split(ret)

    async def write(self, buf: Pointer[T]) -> t.Tuple[Pointer[T], Pointer[T]]:
        """write to a file descriptor

        manpage: write(2)
        """
        self._validate()
        buf.check_address_space(self.task)
        ret = await _write(self.task.sysif, self.near, buf.near, buf.size())
        return buf.split(ret)

class SeekableFileDescriptor(IOFileDescriptor):
    async def pread(self, buf: Pointer[T], offset: int) -> t.Tuple[ReadablePointer[T], Pointer[T]]:
        self._validate()
        with buf.borrow(self.task):
            ret = await _pread(self.task.sysif, self.near, buf.near, buf.size(), offset)
            return buf.readable_split(ret)

    async def pwrite(self, buf: Pointer[T], offset: int) -> t.Tuple[Pointer[T], Pointer[T]]:
        self._validate()
        with buf.borrow(self.task):
            ret = await _pwrite(self.task.sysif, self.near, buf.near, buf.size(), offset)
            return buf.split(ret)

    async def lseek(self, offset: int, whence: SEEK) -> int:
        self._validate()
        return (await _lseek(self.task.sysif, self.near, offset, whence))

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _read(sysif: SyscallInterface, fd: near.FileDescriptor,
                buf: near.Address, count: int) -> int:
    return (await sysif.syscall(SYS.read, fd, buf, count))

async def _write(sysif: SyscallInterface, fd: near.FileDescriptor,
                 buf: near.Address, count: int) -> int:
    return (await sysif.syscall(SYS.write, fd, buf, count))

async def _pread(sysif: SyscallInterface, fd: near.FileDescriptor,
                 buf: near.Address, count: int, offset: int) -> int:
    return (await sysif.syscall(SYS.pread64, fd, buf, count, offset))

async def _pwrite(sysif: SyscallInterface, fd: near.FileDescriptor,
                  buf: near.Address, count: int, offset: int) -> int:
    return (await sysif.syscall(SYS.pwrite64, fd, buf, count, offset))

async def _lseek(sysif: SyscallInterface, fd: near.FileDescriptor,
                 offset: int, whence: SEEK) -> int:
    return (await sysif.syscall(SYS.lseek, fd, offset, whence))



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall.sys.stat import Stat
from rsyscall import local_thread

class TestSeek(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.file = await self.thr.task.memfd_create(await self.thr.ptr("file"))

    async def asyncTearDown(self) -> None:
        await self.file.close()

    async def test_preadwrite(self) -> None:
        stat = await (await self.file.fstat(await self.thr.malloc(Stat))).read()
        self.assertEqual(stat.size, 0)
        read, _ = await self.file.pread(await self.thr.malloc(bytes, 16), offset=0)
        self.assertEqual(read.size(), 0)
        read, _ = await self.file.pread(await self.thr.malloc(bytes, 16), offset=1)
        self.assertEqual(read.size(), 0)
        # we can write to an offset past the end
        data = b'abc'
        wrote, _ = await self.file.pwrite(await self.thr.ptr(data), offset=1)
        self.assertEqual(wrote.size(), len(data))
        # the data is written fine
        read, _ = await self.file.pread(await self.thr.malloc(bytes, 16), offset=1)
        self.assertEqual(await read.read(), data)
        # the earlier bytes are now zeros
        read, _ = await self.file.pread(await self.thr.malloc(bytes, 16), offset=0)
        self.assertEqual(await read.read(), b'\0' + data)
        # size is now 4
        stat = await (await self.file.fstat(await self.thr.malloc(Stat))).read()
        self.assertEqual(stat.size, 4)



"Central thread class, with helper methods to ease a number of common tasks"
from __future__ import annotations
from dataclasses import dataclass
from rsyscall.command import Command
from rsyscall.environ import Environment
from rsyscall.epoller import Epoller, AsyncFileDescriptor
from rsyscall.handle import FileDescriptor, WrittenPointer, Pointer, Task
from rsyscall.handle.fd import _close
from rsyscall.loader import Trampoline, NativeLoader
from rsyscall.memory.ram import RAM
from rsyscall.monitor import AsyncChildProcess, ChildProcessMonitor
from rsyscall.network.connection import Connection
from rsyscall.path import Path
from rsyscall.struct import FixedSize, T_fixed_size, HasSerializer, T_has_serializer, FixedSerializer, T_fixed_serializer, T_pathlike
from rsyscall.tasks.clone import clone_child_task
import logging
import os
import rsyscall.near
import rsyscall.near.types as near
import trio
import typing as t

from rsyscall.fcntl import O, F, FD, _fcntl
from rsyscall.linux.dirent import DirentList
from rsyscall.sched import CLONE
from rsyscall.signal import Sigset, SIG, SignalBlock, HowSIG
from rsyscall.sys.mount import MS
from rsyscall.sys.wait import ChildState, W
from rsyscall.sys.socket import Socketpair, AF, SOCK, T_sockaddr, Sockbuf
from rsyscall.unistd import Pipe, ArgList

logger = logging.getLogger(__name__)

async def write_user_mappings(thr: Thread, uid: int, gid: int,
                              in_namespace_uid: int=None, in_namespace_gid: int=None) -> None:
    """Set up a new user namespace with single-user {uid,gid}_map

    These are the only valid mappings for unprivileged user namespaces.
    """
    if in_namespace_uid is None:
        in_namespace_uid = uid
    if in_namespace_gid is None:
        in_namespace_gid = gid
    procself = Path("/proc/self")

    uid_map = await thr.task.open(await thr.ram.ptr(procself/"uid_map"), O.WRONLY)
    await uid_map.write(await thr.ram.ptr(f"{in_namespace_uid} {uid} 1\n".encode()))
    await uid_map.close()

    setgroups = await thr.task.open(await thr.ram.ptr(procself/"setgroups"), O.WRONLY)
    await setgroups.write(await thr.ram.ptr(b"deny"))
    await setgroups.close()

    gid_map = await thr.task.open(await thr.ram.ptr(procself/"gid_map"), O.WRONLY)
    await gid_map.write(await thr.ram.ptr(f"{in_namespace_gid} {gid} 1\n".encode()))
    await gid_map.close()

async def do_cloexec_except(thr: Thread, excluded_fds: t.Set[near.FileDescriptor]) -> None:
    "Close all CLOEXEC file descriptors, except for those in a whitelist. Would be nice to have a syscall for this."
    # it's important to do this so we can't try to inherit the fds that we close here
    thr.task.fd_table.remove_inherited()
    buf = await thr.ram.malloc(DirentList, 4096)
    dirfd = await thr.task.open(await thr.ram.ptr("/proc/self/fd"), O.DIRECTORY)
    excluded_fds.add(dirfd.near)
    async def maybe_close(fd: near.FileDescriptor) -> None:
        flags = await _fcntl(thr.task.sysif, fd, F.GETFD)
        if (flags & FD.CLOEXEC) and (fd not in excluded_fds):
            await _close(thr.task.sysif, fd)
    async with trio.open_nursery() as nursery:
        while True:
            valid, rest = await dirfd.getdents(buf)
            if valid.size() == 0:
                break
            dents = await valid.read()
            for dent in dents:
                try:
                    num = int(dent.name)
                except ValueError:
                    continue
                nursery.start_soon(maybe_close, near.FileDescriptor(num))
            buf = valid.merge(rest)
    await dirfd.close()

class Thread:
    "A central class holding everything necessary to work with some thread, along with various helpers"
    def __init__(self,
                 task: Task,
                 ram: RAM,
                 connection: Connection,
                 loader: NativeLoader,
                 epoller: Epoller,
                 child_monitor: ChildProcessMonitor,
                 environ: Environment,
                 stdin: FileDescriptor,
                 stdout: FileDescriptor,
                 stderr: FileDescriptor,
    ) -> None:
        self.task = task
        "The `Task` associated with this process"
        self.ram = ram
        self.epoller = epoller
        self.connection = connection
        "This thread's `rsyscall.network.connection.Connection`"
        self.loader = loader
        self.monitor = child_monitor
        self.environ = environ
        "This thread's `rsyscall.environ.Environment`"
        self.stdin = stdin
        "The standard input `FileDescriptor` (FD 0)"
        self.stdout = stdout
        "The standard output `FileDescriptor` (FD 1)"
        self.stderr = stderr
        "The standard error `FileDescriptor` (FD 2)"

    def _init_from(self, thr: Thread) -> None:
        self.task = thr.task
        self.ram = thr.ram
        self.epoller = thr.epoller
        self.connection = thr.connection
        self.loader = thr.loader
        self.monitor = thr.monitor
        self.environ = thr.environ
        self.stdin = thr.stdin
        self.stdout = thr.stdout
        self.stderr = thr.stderr

    @t.overload
    async def malloc(self, cls: t.Type[T_fixed_size]) -> Pointer[T_fixed_size]:
        "malloc a fixed size type"
        pass
    @t.overload
    async def malloc(self, cls: t.Type[T_fixed_serializer], size: int) -> Pointer[T_fixed_serializer]:
        "malloc specifying a specific size"
        pass
    @t.overload
    async def malloc(self, cls: t.Type[T_pathlike], size: int) -> Pointer[T_pathlike]: ...
    @t.overload
    async def malloc(self, cls: t.Type[str], size: int) -> Pointer[str]: ...
    @t.overload
    async def malloc(self, cls: t.Type[bytes], size: int) -> Pointer[bytes]: ...

    async def malloc(self, cls: t.Type[t.Union[FixedSize, FixedSerializer, os.PathLike, str, bytes]],
                     size: int=None) -> Pointer:
        """Allocate a buffer for this type, with the specified size if required

        If `malloc` is given a `rsyscall.struct.FixedSize` type, the size must not be passed;
        if `malloc` is given any other type, the size must be passed.

        Any type which inherits from `rsyscall.struct.FixedSerializer` is supported.
        As a special case for convenience, `bytes`, `str`, and `os.PathLike` are also supported;
        `bytes` will be written out as they are, and `str` and `os.PathLike` will be null-terminated.

        """
        return await self.ram.malloc(cls, size) # type: ignore

    @t.overload
    async def ptr(self, data: T_has_serializer) -> WrittenPointer[T_has_serializer]: ...
    @t.overload
    async def ptr(self, data: T_pathlike) -> WrittenPointer[T_pathlike]: ...
    @t.overload
    async def ptr(self, data: str) -> WrittenPointer[str]: ...
    @t.overload
    async def ptr(self, data: bytes) -> WrittenPointer[bytes]: ...
    async def ptr(self, data: t.Union[HasSerializer, os.PathLike, str, bytes]) -> WrittenPointer:
        """Allocate a buffer for this data, and write the data to that buffer

        Any value which inherits from `HasSerializer` is supported.
        As a special case for convenience, `bytes`, `str`, and `os.PathLike` are also supported;
        `bytes` will be written out as they are, and `str` and `os.PathLike` will be null-terminated.
        """
        return await self.ram.ptr(data)

    async def make_afd(self, fd: FileDescriptor, set_nonblock: bool=False) -> AsyncFileDescriptor:
        """Make an AsyncFileDescriptor; make it nonblocking if `set_nonblock` is True.

        Make sure that `fd` is already in non-blocking mode;
        such as by accepting it with the `SOCK.NONBLOCK` flag;
        if it's not, you can pass set_nonblock=True to make it nonblocking.

        """
        if set_nonblock:
            await fd.fcntl(F.SETFL, O.NONBLOCK)
        return await AsyncFileDescriptor.make(self.epoller, self.ram, fd)

    async def open_async_channels(self, count: int) -> t.List[t.Tuple[AsyncFileDescriptor, FileDescriptor]]:
        "Calls self.connection.open_async_channels; see `Connection.open_async_channels`"
        return (await self.connection.open_async_channels(count))

    async def open_channels(self, count: int) -> t.List[t.Tuple[FileDescriptor, FileDescriptor]]:
        "Calls self.connection.open_channels; see `Connection.open_channels`"
        return (await self.connection.open_channels(count))

    @t.overload
    async def spit(self, path: FileDescriptor, text: t.Union[str, bytes]) -> None:
        pass

    @t.overload
    async def spit(self, path: Path, text: t.Union[str, bytes], mode=0o644) -> Path:
        pass

    async def spit(self, path: t.Union[Path, FileDescriptor], text: t.Union[str, bytes], mode=0o644) -> t.Optional[Path]:
        """Open a file, creating and truncating it, and write the passed text to it

        Probably shouldn't use this on FIFOs or anything.

        Returns the passed-in Path so this serves as a nice pseudo-constructor.

        """
        if isinstance(path, Path):
            out: t.Optional[Path] = path
            fd = await self.task.open(await self.ram.ptr(path), O.WRONLY|O.TRUNC|O.CREAT, mode=mode)
        else:
            out = None
            fd = path
        to_write: Pointer = await self.ram.ptr(os.fsencode(text))
        while to_write.size() > 0:
            _, to_write = await fd.write(to_write)
        await fd.close()
        return out

    async def bind_getsockname(self, sock: FileDescriptor, addr: T_sockaddr) -> T_sockaddr:
        """Call bind and then getsockname on `sock`.

        bind followed by getsockname is a common pattern when allocating unused
        source ports with SockaddrIn(0, ...).  Unfortunately, memory allocation
        for getsockname is quite verbose, so it would be nice to have a helper
        to make that pattern easier. Since we don't want to encourage usage of
        getsockname (it should be rarely used outside of that pattern), we add a
        helper for that specific pattern, rather than getsockname on its own.

        """
        written_addr_ptr = await self.ram.ptr(addr)
        await sock.bind(written_addr_ptr)
        sockbuf_ptr = await sock.getsockname(await self.ram.ptr(Sockbuf(written_addr_ptr)))
        addr_ptr = (await sockbuf_ptr.read()).buf
        return await addr_ptr.read()

    async def mkdir(self, path: Path, mode=0o755) -> Path:
        "Make a directory at this path"
        await self.task.mkdir(await self.ram.ptr(path))
        return path

    async def read_to_eof(self, fd: FileDescriptor) -> bytes:
        "Read this file descriptor until we get EOF, then return all the bytes read"
        data = b""
        while True:
            read, rest = await fd.read(await self.ram.malloc(bytes, 4096))
            if read.size() == 0:
                return data
            # TODO this would be more efficient if we batched our memory-reads at the end
            data += await read.read()

    async def mount(self, source: t.Union[str, os.PathLike], target: t.Union[str, os.PathLike],
                    filesystemtype: str, mountflags: MS,
                    data: str) -> None:
        "Call mount with these args"
        async def op(sem: RAM) -> t.Tuple[
                WrittenPointer[t.Union[str, os.PathLike]], WrittenPointer[t.Union[str, os.PathLike]],
                WrittenPointer[str], WrittenPointer[str]]:
            return (
                await sem.ptr(source),
                await sem.ptr(target),
                await sem.ptr(filesystemtype),
                await sem.ptr(data),
            )
        source_ptr, target_ptr, filesystemtype_ptr, data_ptr = await self.ram.perform_batch(op)
        await self.task.mount(source_ptr, target_ptr, filesystemtype_ptr, mountflags, data_ptr)

    async def socket(self, domain: AF, type: SOCK, protocol: int=0) -> FileDescriptor:
        return await self.task.socket(domain, type, protocol)

    async def pipe(self) -> Pipe:
        return await (await self.task.pipe(await self.malloc(Pipe))).read()

    async def socketpair(self, domain: AF, type: SOCK, protocol: int=0) -> Socketpair:
        return await (await self.task.socketpair(domain, type, protocol, await self.malloc(Socketpair))).read()

    async def chroot(self, path: t.Union[str, os.PathLike]) -> None:
        await self.task.chroot(await self.ptr(path))

    def inherit_fd(self, fd: FileDescriptor) -> FileDescriptor:
        return self.task.inherit_fd(fd)

    async def clone(self, flags: CLONE=CLONE.NONE, automatically_write_user_mappings: bool=True) -> ChildThread:
        """Create a new child thread

        manpage: clone(2)
        """
        child_process, task = await clone_child_task(
            self.task, self.ram, self.connection, self.loader, self.monitor,
            flags, lambda sock: Trampoline(self.loader.server_func, [sock, sock]))
        ram = RAM(task,
                  # We don't inherit the transport because it leads to a deadlock:
                  # If when a child task calls transport.read, it performs a syscall in the child task,
                  # then the parent task will need to call waitid to monitor the child task during the syscall,
                  # which will in turn need to also call transport.read.
                  # But the child is already using the transport and holding the lock,
                  # so the parent will block forever on taking the lock,
                  # and child's read syscall will never complete.
                  self.ram.transport,
                  self.ram.allocator.inherit(task),
        )
        if flags & CLONE.NEWPID:
            # if the new process is pid 1, then CLONE_PARENT isn't allowed so we can't use inherit_to_child.
            # if we are a reaper, than we don't want our child CLONE_PARENTing to us, so we can't use inherit_to_child.
            # in both cases we just fall back to making a new ChildProcessMonitor for the child.
            epoller = await Epoller.make_root(ram, task)
            # this signal is already blocked, we inherited the block, um... I guess...
            # TODO handle this more formally
            signal_block = SignalBlock(task, await ram.ptr(Sigset({SIG.CHLD})))
            monitor = await ChildProcessMonitor.make(ram, task, epoller, signal_block=signal_block)
        else:
            epoller = self.epoller.inherit(ram)
            monitor = self.monitor.inherit_to_child(task)
        thread = ChildThread(Thread(
            task, ram,
            self.connection.inherit(task, ram),
            self.loader,
            epoller, monitor,
            self.environ.inherit(task, ram),
            stdin=self.stdin.inherit(task),
            stdout=self.stdout.inherit(task),
            stderr=self.stderr.inherit(task),
        ), child_process)
        if flags & CLONE.NEWUSER and automatically_write_user_mappings:
            # hack, we should really track the [ug]id ahead of this so we don't have to get it
            # we have to get the [ug]id from the parent because it will fail in the child
            uid = await self.task.getuid()
            gid = await self.task.getgid()
            await write_user_mappings(thread, uid, gid)
        return thread

    async def run(self, command: Command, check=True,
                  *, task_status=trio.TASK_STATUS_IGNORED) -> ChildState:
        """Run the passed command to completion and return its end state, throwing if unclean

        If check is False, we won't throw if the end state is unclean.

        """
        thread = await self.clone()
        child = await thread.exec(command)
        task_status.started(child)
        if check:
            return await child.check()
        else:
            return await child.waitpid(W.EXITED)

    async def unshare(self, flags: CLONE) -> None:
        "Call the unshare syscall, appropriately updating values on this class"
        # Note: unsharing NEWPID causes us to not get zombies for our children if init dies. That
        # means we'll get ECHILDs, and various races can happen. It's not possible to robustly
        # unshare NEWPID.
        if flags & CLONE.FILES:
            await self.unshare_files()
            flags ^= CLONE.FILES
        if flags & CLONE.NEWUSER:
            await self.unshare_user()
            flags ^= CLONE.NEWUSER
            if flags & CLONE.FS:
                flags ^= CLONE.FS
        await self.task.unshare(flags)

    async def unshare_files(self, going_to_exec=True) -> None:
        """Unshare the file descriptor table.

        Set going_to_exec to False if you are going to keep this task around long-term, and we'll do
        a manual cloexec in userspace to clear out fds held by any other non-rsyscall libraries,
        which are automatically copied by Linux into the new fd space.

        We default going_to_exec to True because there's little reason to call unshare_files other
        than to then exec; and even if you do want to call unshare_files without execing, there
        probably aren't any significant other libraries in the FD space; and even if there are such
        libraries, it usually doesn't matter to keep stray references around to their FDs.

        TODO maybe this should return an object that lets us unset CLOEXEC on things?

        """
        await self.task.unshare_files()
        if not going_to_exec:
            await do_cloexec_except(self, set([fd.near for fd in self.task.fd_handles]))

    async def unshare_user(self,
                           in_namespace_uid: int=None, in_namespace_gid: int=None) -> None:
        """Unshare the user namespace.

        We automatically set up the user namespace in the unprivileged way using a single user
        mapping line.  You can pass `in_namespace_uid` and `in_namespace_gid` to control what user
        id and group id we'll observe inside the namespace.  If you want further control, call
        task.unshare(CLONE.NEWUSER) directly.

        We also automatically do unshare(CLONE.FS); that's required by CLONE.NEWUSER.

        """
        uid = await self.task.getuid()
        gid = await self.task.getgid()
        await self.task.unshare(CLONE.FS|CLONE.NEWUSER)
        await write_user_mappings(self, uid, gid,
                                  in_namespace_uid=in_namespace_uid, in_namespace_gid=in_namespace_gid)

    async def exit(self, status: int=0) -> None:
        """Exit this thread

        Currently we just forward through to exit the task.

        I feel suspicious that this method will at some point require more heavy lifting with
        namespaces and monitored children, so I'm leaving it on Thread to prepare for that
        eventuality.

        manpage: exit(2)
        """
        await self.task.exit(status)

    def __repr__(self) -> str:
        name = type(self).__name__
        return f'{name}({self.task})'

class ChildThread(Thread):
    "A thread that we know is also a direct child process of another thread"
    def __init__(self, thr: Thread, process: AsyncChildProcess) -> None:
        super()._init_from(thr)
        self.process = process

    async def _execve(self, path: t.Union[str, os.PathLike], argv: t.List[str], envp: t.List[str],
                      command: Command=None,
    ) -> AsyncChildProcess:
        "Call execve, abstracting over memory; self.{exec,execve} are probably preferable"
        async def op(sem: RAM) -> t.Tuple[WrittenPointer[t.Union[str, os.PathLike]],
                                          WrittenPointer[ArgList],
                                          WrittenPointer[ArgList]]:
            argv_ptrs = ArgList([await sem.ptr(arg) for arg in argv])
            envp_ptrs = ArgList([await sem.ptr(arg) for arg in envp])
            return (await sem.ptr(path),
                    await sem.ptr(argv_ptrs),
                    await sem.ptr(envp_ptrs))
        filename, argv_ptr, envp_ptr = await self.ram.perform_batch(op)
        await self.task.execve(filename, argv_ptr, envp_ptr, command=command)
        return self.process

    async def execv(self, path: t.Union[str, os.PathLike],
                    argv: t.Sequence[t.Union[str, os.PathLike]],
                    command: Command=None,
    ) -> AsyncChildProcess:
        """Replace the running executable in this thread with another; see execve.
        """
        async def op(sem: RAM) -> t.Tuple[WrittenPointer[t.Union[str, os.PathLike]], WrittenPointer[ArgList]]:
            argv_ptrs = ArgList([await sem.ptr(arg) for arg in argv])
            return (await sem.ptr(path), await sem.ptr(argv_ptrs))
        filename_ptr, argv_ptr = await self.ram.perform_batch(op)
        envp_ptr = await self.environ.as_arglist(self.ram)
        await self.task.execve(filename_ptr, argv_ptr, envp_ptr, command=command)
        return self.process

    async def execve(self, path: t.Union[str, os.PathLike],
                     argv: t.Sequence[t.Union[str, os.PathLike]],
                     env_updates: t.Mapping[str, t.Union[str, os.PathLike]]={},
                     inherited_signal_blocks: t.List[SignalBlock]=[],
                     command: Command=None,
    ) -> AsyncChildProcess:
        """Replace the running executable in this thread with another.

        self.exec is probably preferable; it takes a nice Command object which
        is easier to work with.

        We take inherited_signal_blocks as an argument so that we can default it
        to "inheriting" an empty signal mask. Most programs expect the signal
        mask to be cleared on startup. Since we're using signalfd as our signal
        handling method, we need to block signals with the signal mask; and if
        those blocked signals were inherited across exec, other programs would
        break (SIGCHLD is the most obvious example).

        We could depend on the user clearing the signal mask before calling
        exec, similar to how we require the user to remove CLOEXEC from
        inherited fds; but that is a fairly novel requirement to most, so for
        simplicity we just default to clearing the signal mask before exec, and
        allow the user to explicitly pass down additional signal blocks.

        """
        sigmask: t.Set[SIG] = set()
        for block in inherited_signal_blocks:
            sigmask = sigmask.union(block.mask)
        await self.task.sigprocmask((HowSIG.SETMASK, await self.ram.ptr(Sigset(sigmask))))
        if not env_updates:
            # use execv if we aren't updating the env, as an optimization.
            return await self.execv(path, argv, command=command)
        envp: t.Dict[str, str] = {**self.environ.data}
        for key, value in env_updates.items():
            envp[key] = os.fsdecode(value)
        raw_envp: t.List[str] = ['='.join([key, value]) for key, value in envp.items()]
        logger.debug("execveat(%s, %s, %s)", path, argv, env_updates)
        return await self._execve(path, [os.fsdecode(arg) for arg in argv], raw_envp, command=command)

    async def exec(self, command: Command,
                   inherited_signal_blocks: t.List[SignalBlock]=[],
    ) -> AsyncChildProcess:
        """Replace the running executable in this thread with what's specified in `command`

        See self.execve's docstring for an explanation of inherited_signal_blocks.

        manpage: execve(2)
        """
        return (await self.execve(command.executable_path, command.arguments, command.env_updates,
                                  inherited_signal_blocks=inherited_signal_blocks, command=command))



from dataclasses import dataclass
from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread, FileDescriptor, Thread
from rsyscall.fcntl import O
from rsyscall.sys.socket import AF, SOCK, SO, SOL, Sockbuf
from rsyscall.net.if_ import *
from rsyscall.netinet.ip import SockaddrIn
from rsyscall.linux.netlink import *
from rsyscall.linux.rtnetlink import *
from pyroute2 import IPBatch
from rsyscall.sched import CLONE
from rsyscall.thread import write_user_mappings
import errno
import ipaddress
import logging
import trio
logger = logging.getLogger(__name__)
from rsyscall.struct import Int32

@dataclass
class Tun:
    thr: Thread
    fd: FileDescriptor
    name: str
    addr: ipaddress.IPv4Address
    sock: FileDescriptor

    @classmethod
    async def make(cls, parent: Thread, addr: ipaddress.IPv4Address, peer: ipaddress.IPv4Address) -> None:
        # put each tun in a separate netns; I tried putting them in
        # the same netns, but got silent packet delivery failures.
        thr = await parent.clone(CLONE.NEWNET|CLONE.FILES)
        self = cls(
            thr,
            await thr.task.open(await thr.ptr("/dev/net/tun"), O.RDWR),
            'tun0',
            addr,
            await thr.task.socket(AF.INET, SOCK.STREAM),
        )
        await self.fd.ioctl(TUNSETIFF, await self.thr.ptr(Ifreq(self.name, flags=IFF.TUN|IFF.NO_PI)))
        # set up the tun - these ioctls don't actually affect the socket
        await self.sock.ioctl(SIOC.SIFFLAGS, await self.thr.ptr(Ifreq(self.name, flags=IFF.UP)))
        await self.sock.ioctl(SIOC.SIFADDR, await self.thr.ptr(Ifreq(self.name, addr=SockaddrIn(0, self.addr))))
        await self.sock.ioctl(SIOC.SIFDSTADDR, await self.thr.ptr(Ifreq(self.name, addr=SockaddrIn(0, peer))))
        await self.sock.bind(await self.thr.ptr(SockaddrIn(1234, self.addr)))
        return self

class TestNet(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = await local_thread.clone(CLONE.NEWUSER, automatically_write_user_mappings=False)
        await write_user_mappings(
            self.thr,
            await local_thread.task.getuid(), await local_thread.task.getgid(),
            0, 0,
        )

    async def asyncTearDown(self) -> None:
        await self.thr.exit(0)

    async def test_setns_ownership(self) -> None:
        netnsfd = await self.thr.task.open(await self.thr.ram.ptr("/proc/self/ns/net"), O.RDONLY)
        thread = await self.thr.clone(CLONE.FILES)
        await thread.unshare_user()
        with self.assertRaises(PermissionError):
            # we can't setns to a namespace that we don't own, which is fairly lame
            await thread.task.setns(netnsfd, CLONE.NEWNET)

    async def test_ioctl(self) -> None:
        await self.thr.unshare(CLONE.NEWNET)
        tun_fd = await self.thr.task.open(await self.thr.ram.ptr("/dev/net/tun"), O.RDWR)
        name = 'tun0'
        ptr = await self.thr.ram.ptr(Ifreq(name, flags=IFF.TUN))
        await tun_fd.ioctl(TUNSETIFF, ptr)
        sock = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await sock.ioctl(SIOC.GIFINDEX, ptr)
        # this is the second interface in an empty netns
        self.assertEqual((await ptr.read()).ifindex, 2)
        # set it up
        ptr = await ptr.write(Ifreq(name, flags=IFF.UP))
        await sock.ioctl(SIOC.SIFFLAGS, ptr)
        await sock.ioctl(SIOC.GIFFLAGS, ptr)
        self.assertIn(IFF.UP, (await ptr.read()).flags) # type: ignore
        # add IP address
        addr = SockaddrIn(0, '10.0.0.1')
        ptr = await ptr.write(Ifreq(name, addr=addr))
        await sock.ioctl(SIOC.SIFADDR, ptr)
        await sock.ioctl(SIOC.GIFADDR, ptr)
        self.assertEqual(addr, (await ptr.read()).addr.parse())

    async def test_rtnetlink(self) -> None:
        await self.thr.unshare(CLONE.NEWNET)
        netsock = await self.thr.task.socket(AF.NETLINK, SOCK.DGRAM, NETLINK.ROUTE)
        await netsock.bind(await self.thr.ram.ptr(SockaddrNl(0, RTMGRP.LINK)))

        tun_fd = await self.thr.task.open(await self.thr.ram.ptr("/dev/net/tun"), O.RDWR)
        ptr = await self.thr.ram.ptr(Ifreq('tun0', flags=IFF.TUN))
        await tun_fd.ioctl(TUNSETIFF, ptr)
        sock = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await sock.ioctl(SIOC.GIFINDEX, ptr)
        # this is the second interface in an empty netns
        self.assertEqual((await ptr.read()).ifindex, 2)

        valid, _ = await netsock.read(await self.thr.ram.malloc(bytes, 4096))
        batch = IPBatch()
        evs = batch.marshal.parse(await valid.read())
        self.assertEqual(len(evs), 1)
        self.assertEqual(evs[0]['event'], 'RTM_NEWLINK')

    async def test_connected_tun(self) -> None:
        left_addr, right_addr = [ipaddress.IPv4Address(f'10.0.0.{i}') for i in [1, 2]]
        left = await Tun.make(self.thr, left_addr, right_addr)
        right = await Tun.make(self.thr, right_addr, left_addr)

        logger.info("Use socat to connect the two tuns bidirectionally")
        socat = await self.thr.environ.which('socat')
        socat_thr = await self.thr.clone()
        await socat_thr.inherit_fd(left.fd).dup2(socat_thr.stdin)
        await socat_thr.inherit_fd(right.fd).dup2(socat_thr.stdout)
        socat_proc = await socat_thr.exec(socat.args('STDIN', 'STDOUT'))

        conn, acc = left, right
        logger.info("%s will connect and %s will accept", conn, acc)
        await acc.sock.listen(1)
        logger.info("Run connect and accept in parallel; this works even though we aren't using AFDs, "
                    "because the two sockets are in separate threads.")
        async with trio.open_nursery() as nursery:
            # TODO this blocks the thread for some reason...
            # nursery.start_soon(socat_proc.check)
            @nursery.start_soon
            async def do_conn():
                await conn.sock.connect(await conn.thr.ptr(SockaddrIn(1234, acc.addr)))
            accepted_conn = await acc.sock.accept()
        await conn.sock.write(await self.thr.ptr(b'hello world'))
        await accepted_conn.read(await self.thr.malloc(bytes, 4096))

class TestNetLocalPort(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = await local_thread.clone(CLONE.NEWUSER, automatically_write_user_mappings=False)
        await write_user_mappings(
            self.thr,
            await local_thread.task.getuid(), await local_thread.task.getgid(),
            0, 0,
        )
        await self.thr.unshare(CLONE.NEWNET)
        local = await self.thr.task.open(await self.thr.ram.ptr("/proc/sys/net/ipv4/ip_local_port_range"), O.WRONLY)
        await local.write(await self.thr.ptr("40000 40000\n"))
        await local.close()

        self.ioctl_sock = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await self.ioctl_sock.ioctl(SIOC.SIFFLAGS, await self.thr.ptr(Ifreq("lo", flags=IFF.UP)))

    async def asyncTearDown(self) -> None:
        await self.thr.exit(0)

    async def test_reuseaddr_listen(self) -> None:
        """If you use SO.REUSEADDR, your local and peer address can be the same

        This is kind of alarming and surprising, but it's a real behavior.
        """
        sockfd = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await sockfd.setsockopt(SOL.SOCKET, SO.REUSEADDR, await self.thr.ptr(Int32(1)))
        addr = await self.thr.bind_getsockname(sockfd, SockaddrIn(0, '127.0.0.1'))

        sockfd2 = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await sockfd2.setsockopt(SOL.SOCKET, SO.REUSEADDR, await self.thr.ptr(Int32(1)))
        await sockfd2.bind(await self.thr.ptr(addr))
        await sockfd.listen(10)

        await sockfd2.connect(await self.thr.ptr(addr))

        sockbuf_ptr = await sockfd2.getsockname(await self.thr.ptr(Sockbuf(await self.thr.malloc(SockaddrIn))))
        self.assertEqual(addr, await (await sockbuf_ptr.read()).buf.read())
        sockbuf_ptr = await sockfd2.getpeername(await self.thr.ptr(Sockbuf(await self.thr.malloc(SockaddrIn))))
        self.assertEqual(addr, await (await sockbuf_ptr.read()).buf.read())

    async def test_stream_reuseaddr(self) -> None:
        "With STREAM sockets, even if you use SO.REUSEADDR, binding 0 twice will never give you the same port."
        sockfd = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await sockfd.setsockopt(SOL.SOCKET, SO.REUSEADDR, await self.thr.ptr(Int32(1)))
        await sockfd.bind(await self.thr.ram.ptr(SockaddrIn(0, '127.0.0.1')))

        sockfd2 = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await sockfd2.setsockopt(SOL.SOCKET, SO.REUSEADDR, await self.thr.ptr(Int32(1)))
        with self.assertRaises(OSError) as cm:
            await sockfd2.bind(await self.thr.ram.ptr(SockaddrIn(0, '127.0.0.1')))
        self.assertEqual(cm.exception.errno, errno.EADDRINUSE)

    async def test_dgram(self) -> None:
        "With DGRAM sockets, without SO.REUSEADDR, binding 0 twice will never give you the same port."
        sockfd = await self.thr.task.socket(AF.INET, SOCK.DGRAM)
        await sockfd.bind(await self.thr.ram.ptr(SockaddrIn(0, '127.0.0.1')))

        sockfd2 = await self.thr.task.socket(AF.INET, SOCK.DGRAM)
        with self.assertRaises(OSError) as cm:
            await sockfd2.bind(await self.thr.ram.ptr(SockaddrIn(0, '127.0.0.1')))
        self.assertEqual(cm.exception.errno, errno.EADDRINUSE)

    async def test_dgram_reuseaddr(self) -> None:
        """With DGRAM sockets, if you use SO.REUSEADDR, binding 0 *can* give you the same port.

        But note that you can also just set REUSEADDR after binding.
        """
        sockfd = await self.thr.task.socket(AF.INET, SOCK.DGRAM)
        await sockfd.setsockopt(SOL.SOCKET, SO.REUSEADDR, await self.thr.ptr(Int32(1)))
        addr = await self.thr.bind_getsockname(sockfd, SockaddrIn(0, '127.0.0.1'))

        sockfd2 = await self.thr.task.socket(AF.INET, SOCK.DGRAM)
        await sockfd2.setsockopt(SOL.SOCKET, SO.REUSEADDR, await self.thr.ptr(Int32(1)))
        addr2 = await self.thr.bind_getsockname(sockfd2, SockaddrIn(0, '127.0.0.1'))

        self.assertEqual(addr, addr2)



"""`#include <sys/ioctl.h>`

Abandon all hope, ye who enter here.

"""
import typing as t
import errno

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor
from rsyscall.handle.pointer import Pointer

class IoctlFileDescriptor(BaseFileDescriptor):
    async def ioctl(self, request: int, arg: Pointer) -> int:
        self._validate()
        arg._validate()
        try:
            return (await _ioctl(self.task.sysif, self.near, request, arg.near))
        except OSError as e:
            if e.errno == errno.ENOTTY:
                e.filename = request
            raise

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _ioctl(sysif: SyscallInterface, fd: near.FileDescriptor,
                request: int, arg: t.Optional[t.Union[int, near.Address]]=None) -> int:
    if arg is None:
        arg = 0
    return (await sysif.syscall(SYS.ioctl, fd, request, arg))




"""The rsyscall protocol implementation

That's a bit grandiose, because the rsyscall protocol is extremely simple: Write
a syscall request out as a fixed-size struct containing a syscall number and the
arguments, read the syscall response in as a 64-bit long.

Still, it's not completely trivial, because we do pipelining of syscall
requests, and we also batch together multiple requests so they can be written
out all at once.

"""
from rsyscall._raw import ffi # type: ignore
from dataclasses import dataclass
from dneio import RequestQueue, reset, is_running_directly_under_trio
from rsyscall.epoller import AsyncFileDescriptor, AsyncReadBuffer
from rsyscall.handle import Pointer, FileDescriptor
from rsyscall.near.sysif import SyscallHangup, SyscallSendError, SyscallInterface, Syscall, raise_if_error
from rsyscall.struct import Struct, StructList
from rsyscall.sys.socket import SHUT
from rsyscall.sys.syscall import SYS
import logging
import trio
import typing as t

__all__ = [
    "SyscallConnection",
    "RsyscallSyscall",
    "SyscallResponse",
]

class RsyscallSyscall(Struct, Syscall):
    "The struct representing a syscall request"
    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct rsyscall_syscall const*', {
            "sys": self.number,
            "args": (self.arg1, self.arg2, self.arg3, self.arg4, self.arg5, self.arg6),
        })))

    T = t.TypeVar('T', bound='Syscall')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct rsyscall_syscall*', ffi.from_buffer(data))
        return cls(struct.sys,
                   struct.args[0], struct.args[1], struct.args[2],
                   struct.args[3], struct.args[4], struct.args[5])

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct rsyscall_syscall')

@dataclass
class SyscallResponse(Struct):
    "The struct representing a syscall response"
    value: int

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('long const*', self.value)))

    T = t.TypeVar('T', bound='SyscallResponse')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('long*', ffi.from_buffer(data))
        return cls(struct[0])

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('long')

class SyscallConnection(SyscallInterface):
    "A connection to some rsyscall server where we can make syscalls"
    def __init__(self,
                 logger: logging.Logger,
                 tofd: AsyncFileDescriptor,
                 fromfd: AsyncFileDescriptor,
                 server_infd: FileDescriptor,
                 server_outfd: FileDescriptor,
    ) -> None:
        self.logger = logger
        self.tofd = tofd
        self.fromfd = fromfd
        self.server_infd = server_infd
        self.server_outfd = server_outfd
        self.valid: t.Optional[Pointer[bytes]] = None
        self.request_queue = RequestQueue[RsyscallSyscall, int]()
        reset(self._run_requests())
        self.response_queue = RequestQueue[RsyscallSyscall, int]()
        reset(self._run_responses())

    def get_activity_fd(self) -> FileDescriptor:
        """Return an fd which is readable when there's other syscalls waiting to be done

        This is true by definition: this fd is read by the rsyscall server to receive
        syscalls, and when this fd is readable, it means there's syscalls to be read.

        """
        return self.server_infd

    async def close_interface(self) -> None:
        """Close this SyscallConnection; pending requests will throw SyscallHangup

        We don't close server_infd and server_outfd, because we don't have any way to close
        them; they were handles that used this syscall interface, so now they're broken.

        """
        await self.tofd.handle.shutdown(SHUT.RDWR)
        await self.fromfd.handle.shutdown(SHUT.RDWR)

    async def syscall(self, number: SYS, arg1=0, arg2=0, arg3=0, arg4=0, arg5=0, arg6=0) -> int:
        syscall = RsyscallSyscall(number, arg1, arg2, arg3, arg4, arg5, arg6)
        self.logger.debug("%s", syscall)
        try:
            result = await self.do_syscall(syscall)
            raise_if_error(result)
        except OSError as exn:
            self.logger.debug("%s -> %s", number, exn)
            raise OSError(exn.errno, exn.strerror) from None
        except Exception as exn:
            self.logger.debug("%s -/ %s", number, exn)
            raise
        else:
            self.logger.debug("%s -> %s", number, result)
            return result

    async def do_syscall(self, syscall: RsyscallSyscall) -> int:
        """Write a syscall request and perform it.

        """
        # TODO as a hack, so we don't have to figure it out now, we don't allow
        # a syscall request to be cancelled before it's actually made. we could
        # make this work later, and that would reduce some blocking from waitid
        if is_running_directly_under_trio():
            with trio.CancelScope(shield=True):
                # hmm this cancel scope shields the entire thing. unfortunate...
                return await self.request_queue.request(syscall)
        else:
            return await self.request_queue.request(syscall)

    async def _run_requests(self) -> None:
        while True:
            # wait until we have a batch to do, received from self.pending_requests
            requests = await self.request_queue.get_many()
            self.logger.debug("_run_requests: get_many: %s", requests)
            # write remaining_reqs to memory
            ptr: Pointer[StructList] = await self.tofd.ram.ptr(
                StructList(RsyscallSyscall, [syscall for syscall, coro in requests]))
            self.logger.debug("_run_requests: performed ptr for: %s", requests)
            ptr_to_write, reqs_to_write = ptr, requests
            # TODO write requests to tofd in parallel with receiving more
            # requests from the channel and writing them to memory
            try:
                while ptr_to_write.size() > 0:
                    _, ptr_to_write = await self.tofd.write(ptr_to_write)
                    # TODO mark the requests as complete incrementally,
                    # so if we do have a partial write,
                    # we don't block earlier requests on later ones.
            except OSError as syscall_error:
                exn = SyscallSendError()
                exn.__cause__ = syscall_error
                # TODO not necessarily all of the syscalls have failed...
                # some maybe have been actually written, if we had a partial write
                for syscall, cb in reqs_to_write:
                    cb.throw(exn)
            else:
                for syscall, coro in reqs_to_write:
                    self.logger.debug("forward_request: %s", syscall)
                    self.response_queue.request_cb(syscall, coro)

    async def _run_responses(self) -> None:
        buffer = AsyncReadBuffer(self.fromfd)
        while True:
            syscall, cb = await self.response_queue.get_one()
            self.logger.debug("going to read_result for syscall: %s %s", syscall, self.fromfd.handle.near)
            try:
                value = (await buffer.read_struct(SyscallResponse)).value
            except Exception as exn:
                hangup_exn = SyscallHangup()
                hangup_exn.__cause__ = exn
                cb.throw(hangup_exn)
            else:
                cb.send(value)



"""Classes which own resources and provide the main syscall interfaces

We have several resource-owning classes in this module: FileDescriptor, Pointer, Process,
MemoryMapping, etc.

In the analogy to near and far pointers, they are like a near pointer plus a segment
register. A more useful analogy is to "handles" from classic Mac OS/PalmOS/16-bit Windows
memory management. Like handles, these classes are locked on use with the "borrow" context
manager, and they are weakly "relocatable", in that they continue to be valid as the
task's segment ids (namespaces) change. See:
https://en.wikipedia.org/wiki/Mac_OS_memory_management

However, unlike the MacOS handles that are the origin of the name of this module, these
resource-owning classes are garbage collected. Garbage collection should be relied on and
preferred over context managers or explicit closing, which are both far too inflexible for
large scale resource management.

"""
from __future__ import annotations
from rsyscall._raw import ffi, lib # type: ignore
from dataclasses import dataclass
import rsyscall.far
import rsyscall.near
from rsyscall.far import File
import os
import typing as t
import logging
import contextlib
from rsyscall.command import Command
from rsyscall.handle.fd import FileDescriptorTask, BaseFileDescriptor, FDTable
from rsyscall.handle.pointer import Pointer, WrittenPointer, ReadablePointer, LinearPointer
from rsyscall.handle.process import Process, ChildProcess, ThreadProcess, ProcessTask
from rsyscall.near.sysif import UnusableSyscallInterface
logger = logging.getLogger(__name__)

from rsyscall.sched import CLONE, Stack, _unshare
from rsyscall.signal import Siginfo
from rsyscall.fcntl import AT, F, FD
from rsyscall.path import Path
from rsyscall.unistd import SEEK, ArgList, Pipe, OK
from rsyscall.unistd.exec import _execve, _execveat, _exit
from rsyscall.linux.futex import RobustListHead, FutexNode
from rsyscall.sys.wait import W

from rsyscall.fcntl        import               FcntlFileDescriptor
from rsyscall.sys.eventfd  import EventfdTask,  EventFileDescriptor
from rsyscall.sys.timerfd  import TimerfdTask,  TimerFileDescriptor
from rsyscall.sys.epoll    import EpollTask,    EpollFileDescriptor
from rsyscall.sys.inotify  import InotifyTask,  InotifyFileDescriptor
from rsyscall.sys.signalfd import SignalfdTask, SignalFileDescriptor
from rsyscall.sys.mman     import MemoryMappingTask, MappableFileDescriptor
from rsyscall.sys.stat     import               StatFileDescriptor
from rsyscall.signal       import SignalTask
from rsyscall.sys.socket   import SocketTask,   SocketFileDescriptor
from rsyscall.sys.ioctl    import               IoctlFileDescriptor
from rsyscall.linux.dirent import               GetdentsFileDescriptor
from rsyscall.linux.futex  import FutexTask
from rsyscall.linux.memfd  import MemfdTask
from rsyscall.sys.uio      import               UioFileDescriptor
from rsyscall.unistd       import FSTask,       FSFileDescriptor
from rsyscall.unistd.pipe  import PipeTask
from rsyscall.unistd.cwd   import CWDTask
from rsyscall.unistd.credentials import CredentialsTask
from rsyscall.unistd.io    import IOFileDescriptor, SeekableFileDescriptor
from rsyscall.sys.capability import CapabilityTask
from rsyscall.sys.prctl    import PrctlTask
from rsyscall.sys.mount    import MountTask
from rsyscall.sys.resource import ResourceTask
from rsyscall.sched        import SchedTask

# re-exported
from rsyscall.sched import Borrowable

__all__ = [
    "FileDescriptor", "FDTable", "BaseFileDescriptor",
    "Pointer", "WrittenPointer", "ReadablePointer", "LinearPointer",
    "Process", "ChildProcess", "ThreadProcess",
    "Task",
]


################################################################################
# FileDescriptor
T = t.TypeVar('T')
@dataclass(eq=False)
class FileDescriptor(
        EventFileDescriptor, TimerFileDescriptor, EpollFileDescriptor,
        InotifyFileDescriptor, SignalFileDescriptor,
        IoctlFileDescriptor, GetdentsFileDescriptor, UioFileDescriptor,
        SeekableFileDescriptor, IOFileDescriptor,
        FSFileDescriptor,
        SocketFileDescriptor,
        MappableFileDescriptor,
        StatFileDescriptor,
        FcntlFileDescriptor,
        BaseFileDescriptor,
):
    """A file descriptor accessed through some `Task`, with FD-based syscalls as methods

    A `FileDescriptor` represents the ability to use some open file through some `Task`.
    When an open file is created by a syscall in some `Task`,
    the syscall will return a `FileDescriptor` which allows accessing that open file through that `Task`.

    A `FileDescriptor` has many methods to make syscalls;
    most syscalls which take a file descriptor as their first argument are present as a method on `FileDescriptor`.
    These syscalls will be made through the `Task` in the FileDescriptor's `task` field.

    Since there are so many syscalls,
    this class is built by inheriting from many other purpose specific `FooFileDescriptor` classes,
    which in turn all inherit from `BaseFileDescriptor`.

    After we have opened the file and performed some operations on it,
    we can call `close` to immediately close the FileDescriptor and free its resources.
    The FileDescriptor will also be automatically closed in the background
    after the FileDescriptor has been garbage collected.
    Garbage collection should be relied on and preferred over context managers or explicit closing,
    which are both too inflexible for large scale resource management.
    Garbage collection is currently run when we change file descriptor tables,
    as well as on-demand if the user calls `FileDescriptorTask.run_fd_table_gc`.

    We can use `inherit` to copy a FileDescriptor into a task which inherited file descriptors from a parent,
    and `for_task` to copy a FileDescriptor into tasks sharing the same file descriptor table.
    We can also use more complicated methods, such as `rsyscall.sys.socket.CmsgSCMRights`,
    to copy file descriptors without inheritance or a shared file descriptor table.

    """
    __slots__ = ()
    task: Task

    def as_proc_path(self) -> Path:
        """Return the /proc/{pid}/fd/{num} path pointing to this FD.

        This should be used with care, but it's sometimes useful for programs
        which accept paths instead of file descriptors.

        """
        pid = self.task.process.near.id
        num = self.near.number
        return Path(f"/proc/{pid}/fd/{num}")

    async def disable_cloexec(self) -> None:
        "Unset the `FD.CLOEXEC` flag so this file descriptor can be inherited"
        # TODO this doesn't make any sense. we shouldn't allow cloexec if there are multiple people in our fd table;
        # whether or not there are multiple handles to the fd is irrelevant.
        if not self.is_only_handle():
            raise Exception("shouldn't disable cloexec when there are multiple handles to this fd")
        await self.fcntl(F.SETFD, 0)

    async def enable_cloexec(self) -> None:
        "Set the `FD.CLOEXEC` flag so this file descriptor can't be inherited"
        await self.fcntl(F.SETFD, FD.CLOEXEC)

    async def as_argument(self) -> int:
        "`disable_cloexec`, then return this `FileDescriptor` as an integer; useful when passing the FD as an argument"
        await self.disable_cloexec()
        return int(self)

    async def __aenter__(self) -> FileDescriptor:
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        await self.close()

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        if self.valid:
            return f"FD({self.task}, {self.near.number})"
        else:
            return f"FD({self.task}, {self.near.number}, valid=False)"


################################################################################
# Task

class Task(
        EventfdTask[FileDescriptor], TimerfdTask[FileDescriptor], EpollTask[FileDescriptor],
        InotifyTask[FileDescriptor], SignalfdTask[FileDescriptor],
        MemfdTask[FileDescriptor],
        FSTask[FileDescriptor],
        SocketTask[FileDescriptor],
        PipeTask,
        MemoryMappingTask, CWDTask,
        FileDescriptorTask[FileDescriptor],
        CapabilityTask, PrctlTask, MountTask,
        CredentialsTask,
        ProcessTask,
        SchedTask,
        ResourceTask,
        FutexTask,
        SignalTask, rsyscall.far.Task,
):
    """A Linux process context under our control, ready for syscalls

    Since there are many different syscalls we could make,
    this class is built by inheriting from many other purpose specific "Task" classes,
    which in turn all inherit from the base `rsyscall.far.Task`.

    This is named after the kernel struct, "struct task", associated with each process.

    """
    def __init__(self,
                 process: t.Union[rsyscall.near.Process, Process],
                 fd_table: FDTable,
                 address_space: rsyscall.far.AddressSpace,
                 pidns: rsyscall.far.PidNamespace,
    ) -> None:
        super().__init__(
            UnusableSyscallInterface(),
            t.cast(rsyscall.near.Process, process), fd_table, address_space, pidns,
        )

    def _file_descriptor_constructor(self, fd: rsyscall.near.FileDescriptor) -> FileDescriptor:
        # for extensibility
        return FileDescriptor(self, fd, True)

    def _make_fresh_address_space(self) -> None:
        self.address_space = rsyscall.far.AddressSpace(self.process.near.id)

    async def unshare(self, flags: CLONE) -> None:
        if flags & CLONE.FILES:
            await self.unshare_files()
            flags ^= CLONE.FILES
        if flags:
            await _unshare(self.sysif, flags)

    async def setns_user(self, fd: FileDescriptor) -> None:
        # can't setns to a user namespace while sharing CLONE_FS
        await self.unshare(CLONE.FS)
        await self.setns(fd, CLONE.NEWUSER)

    async def execveat(self, fd: t.Optional[FileDescriptor],
                       pathname: WrittenPointer[t.Union[str, os.PathLike]],
                       argv: WrittenPointer[ArgList],
                       envp: WrittenPointer[ArgList],
                       flags: AT=AT.NONE,
                       command: Command=None,
    ) -> None:
        with contextlib.ExitStack() as stack:
            if fd:
                fd_n: t.Optional[rsyscall.near.FileDescriptor] = stack.enter_context(fd.borrow(self))
            else:
                fd_n = None
            stack.enter_context(pathname.borrow(self))
            argv.check_address_space(self)
            envp.check_address_space(self)
            for arg in [*argv.value, *envp.value]:
                stack.enter_context(arg.borrow(self))
            self.manipulating_fd_table = True
            try:
                await _execveat(self.sysif, fd_n, pathname.near, argv.near, envp.near, flags)
            except OSError as exn:
                exn.filename = (fd, pathname.value)
                raise
            finally:
                self.manipulating_fd_table = False
            self._make_fresh_fd_table()
            self._make_fresh_address_space()
            if isinstance(self.process, ChildProcess):
                self.process.did_exec(command)
        await self.sysif.close_interface()

    async def execve(self, filename: WrittenPointer[t.Union[str, os.PathLike]],
                     argv: WrittenPointer[ArgList],
                     envp: WrittenPointer[ArgList],
                     command: Command=None,
    ) -> None:
        filename.check_address_space(self)
        argv.check_address_space(self)
        envp.check_address_space(self)
        for arg in [*argv.value, *envp.value]:
            arg.check_address_space(self)
        self.manipulating_fd_table = True
        try:
            await _execve(self.sysif, filename.near, argv.near, envp.near)
        except OSError as exn:
            exn.filename = filename.value
            raise
        self.manipulating_fd_table = False
        self._make_fresh_fd_table()
        self._make_fresh_address_space()
        if isinstance(self.process, ChildProcess):
            self.process.did_exec(command)
        await self.sysif.close_interface()

    async def exit(self, status: int) -> None:
        self.manipulating_fd_table = True
        await _exit(self.sysif, status)
        self.manipulating_fd_table = False
        self._make_fresh_fd_table()
        # close the syscall interface; we don't have to do this since it'll be
        # GC'd, but maybe we want to be tidy in advance.
        await self.sysif.close_interface()



from __future__ import annotations
from rsyscall._raw import ffi, lib # type: ignore
from dataclasses import dataclass
from rsyscall.struct import FixedSize, Serializer
import rsyscall.near.types as near
from rsyscall.handle.fd import FileDescriptorTask
import typing as t
if t.TYPE_CHECKING:
    from rsyscall.handle import FileDescriptor

T_pipe = t.TypeVar('T_pipe', bound='Pipe')
@dataclass
class Pipe(FixedSize):
    "A pair of file descriptors, as written by pipe."
    read: FileDescriptor
    write: FileDescriptor

    def __getitem__(self, idx: int) -> FileDescriptor:
        if idx == 0:
            return self.read
        elif idx == 1:
            return self.write
        else:
            raise IndexError("only index 0 or 1 are valid for Pipe:", idx)

    def __iter__(self) -> t.Iterable[FileDescriptor]:
        return iter([self.read, self.write])

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fdpair')

    @classmethod
    def get_serializer(cls: t.Type[T_pipe], task: FileDescriptorTask[FileDescriptor]) -> Serializer[T_pipe]:
        return PipeSerializer(cls, task)

@dataclass
class PipeSerializer(Serializer[T_pipe]):
    cls: t.Type[T_pipe]
    task: FileDescriptorTask[FileDescriptor]

    def to_bytes(self, pair: T_pipe) -> bytes:
        struct = ffi.new('struct fdpair*', (pair.read, pair.write))
        return bytes(ffi.buffer(struct))

    def from_bytes(self, data: bytes) -> T_pipe:
        struct = ffi.cast('struct fdpair const*', ffi.from_buffer(data))
        def make(n: int) -> FileDescriptor:
            return self.task.make_fd_handle(near.FileDescriptor(int(n)))
        return self.cls(make(struct.first), make(struct.second))

#### Classes ####
from rsyscall.fcntl import O
from rsyscall.handle.pointer import Pointer, LinearPointer

class PipeTask(FileDescriptorTask):
    async def pipe(self, buf: Pointer[Pipe], flags: O=O.NONE) -> LinearPointer[Pipe]:
        """create pipe

        manpage: pipe2(2)
        """
        # TODO we should force the serializer for the pipe to be using this task...
        # otherwise it could get deserialized by a task with which we share memory,
        # but not share file descriptor tables.
        # Maybe we could create the Serializer right here, and discard
        # the passed-in one? That wouldn't allow a different task in
        # the same fd table to receive the handles though.
        with buf.borrow(self):
            await _pipe(self.sysif, buf.near, flags|O.CLOEXEC)
            return buf._linearize()

#### Raw syscalls ####
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _pipe(sysif: SyscallInterface, pipefd: near.Address, flags: O) -> None:
    await sysif.syscall(SYS.pipe2, pipefd, flags)



"`#include <sys/inotify.h>`"
from __future__ import annotations
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.limits import NAME_MAX
from rsyscall.near.types import WatchDescriptor
from rsyscall.struct import Serializable
import typing as t
from dataclasses import dataclass
import enum
import os

__all__ = [
    "InotifyFlag",
    "IN",
    "InotifyEvent",
    "InotifyEventList",
    "InotifyFileDescriptor",
]

class InotifyFlag(enum.IntFlag):
    NONE = 0
    CLOEXEC = lib.IN_CLOEXEC
    NONBLOCK = lib.IN_NONBLOCK

class IN(enum.IntFlag):
    # possible events, specified in inotify_add_watch and returned in struct inotify_event
    ACCESS = lib.IN_ACCESS
    ATTRIB = lib.IN_ATTRIB
    CLOSE_WRITE = lib.IN_CLOSE_WRITE
    CLOSE_NOWRITE = lib.IN_CLOSE_NOWRITE
    CREATE = lib.IN_CREATE
    DELETE = lib.IN_DELETE
    DELETE_SELF = lib.IN_DELETE_SELF
    MODIFY = lib.IN_MODIFY
    MOVE_SELF = lib.IN_MOVE_SELF
    MOVED_FROM = lib.IN_MOVED_FROM
    MOVED_TO = lib.IN_MOVED_TO
    OPEN = lib.IN_OPEN
    # additional options to inotify_add_watch
    DONT_FOLLOW = lib.IN_DONT_FOLLOW
    EXCL_UNLINK = lib.IN_EXCL_UNLINK
    MASK_ADD = lib.IN_MASK_ADD
    ONESHOT = lib.IN_ONESHOT
    ONLYDIR = lib.IN_ONLYDIR
    # additional bits returned in struct inotify_event 
    IGNORED = lib.IN_IGNORED
    ISDIR = lib.IN_ISDIR
    Q_OVERFLOW = lib.IN_Q_OVERFLOW
    UNMOUNT = lib.IN_UNMOUNT

@dataclass
class InotifyEvent:
    wd: WatchDescriptor
    mask: IN
    cookie: int
    name: t.Optional[str]
    MINIMUM_SIZE_TO_READ_ONE_EVENT = ffi.sizeof('struct inotify_event') + NAME_MAX + 1

    def to_bytes(self) -> bytes:
        if self.name is not None:
            name = self.name.encode()
            name_len = len(name)
        else:
            name = b""
            name_len = 0
        return bytes(ffi.buffer(ffi.new('struct inotify_event*', {
            "wd": self.wd,
            "mask": self.mask,
            "cookie": self.cookie,
            "len": name_len,
            "name": name,
        })))

    T = t.TypeVar('T', bound='InotifyEvent')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> t.Tuple[T, int]:
        struct = ffi.cast('struct inotify_event*', ffi.from_buffer(data))
        value = cls(
            wd=WatchDescriptor(struct.wd),
            mask=IN(struct.mask),
            cookie=struct.cookie,
            name=ffi.string(struct.name, struct.len).decode() if struct.len else None,
        )
        size = ffi.sizeof("struct inotify_event") + struct.len
        return value, size

class InotifyEventList(t.List[InotifyEvent], Serializable):
    def to_bytes(self) -> bytes:
        ret = b""
        for ent in self:
            ret += ent.to_bytes()
        return ret

    T = t.TypeVar('T', bound='InotifyEventList')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        entries = []
        while len(data) > 0:
            ent, size = InotifyEvent.from_bytes(data)
            entries.append(ent)
            data = data[size:]
        return cls(entries)

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask
from rsyscall.handle.pointer import Pointer, WrittenPointer

T_fd = t.TypeVar('T_fd', bound='InotifyFileDescriptor')
class InotifyFileDescriptor(BaseFileDescriptor):
    async def inotify_add_watch(self,
                                pathname: WrittenPointer[t.Union[str, os.PathLike]], mask: IN) -> WatchDescriptor:
        self._validate()
        with pathname.borrow(self.task):
            return (await _inotify_add_watch(self.task.sysif, self.near, pathname.near, mask))

    async def inotify_rm_watch(self, wd: WatchDescriptor) -> None:
        self._validate()
        await _inotify_rm_watch(self.task.sysif, self.near, wd)

class InotifyTask(FileDescriptorTask[T_fd]):
    async def inotify_init(self, flags: InotifyFlag=InotifyFlag.NONE) -> T_fd:
        return self.make_fd_handle(await _inotify_init(self.sysif, flags|InotifyFlag.CLOEXEC))

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _inotify_init(sysif: SyscallInterface, flags: InotifyFlag) -> near.FileDescriptor:
    return near.FileDescriptor(await sysif.syscall(SYS.inotify_init1, flags))

async def _inotify_add_watch(sysif: SyscallInterface, fd: near.FileDescriptor,
                             pathname: near.Address, mask: IN) -> WatchDescriptor:
    return WatchDescriptor(await sysif.syscall(SYS.inotify_add_watch, fd, pathname, mask))

async def _inotify_rm_watch(sysif: SyscallInterface, fd: near.FileDescriptor,
                            wd: WatchDescriptor) -> None:
    await sysif.syscall(SYS.inotify_rm_watch, fd, wd)



from rsyscall.wish import wish, Wish
import trio
import typing as t

async def main() -> None:
    try:
        raise Exception("um")
    except Exception as exn:
        games = await wish(Wish(t.List[str], "i wish you would tell me ur faverut games right now!!!"), from_exn=exn)
    for game in games:
        print(game, "is fun")
    number = await wish(Wish(int, "i wish you would tell me a number!!!"))
    flavor = await wish(Wish(str, "Sorry for being so rude, spirit. Could you tell me your favorite flavor of pie?"))
    pies = [f"A tasty {flavor} pie."]*number
    await wish(Wish(None, f"Here you go spirit! {number} delicious {flavor} pies! Return when you're done eating them!"))
    print("Bye spirit! See you later!")

if __name__ == "__main__":
    trio.run(main)



"A slightly improved version of `pathlib.PurePosixPath`"
from __future__ import annotations
import pathlib

class Path(pathlib.PurePosixPath):
    """A version of `pathlib.PurePosixPath` which is safe to inherit from

    `pathlib` does a lot of crazy stuff which makes it hard to inherit from.  This
    class insulates us from that stuff, so it can be inherited from naively.

    We use this as Path, rather than using `pathlib.Path`, to avoid confusion about
    `pathlib.Path`'s filesystem-interaction methods, which are not rsyscall-aware.

    """
    def __new__(cls, *args, **kwargs) -> Path:
        """Override `pathlib.PurePath.__new__` to restore default behavior

        `pathlib.PurePath` inherits from `object`, so we just use `object.__new__`.
        """
        return object.__new__(cls)

    def __init__(self, *args) -> None:
        """Override `pathlib.PurePath.__init__` to create more sane behavior

        We copy a small amount of code from `pathlib.PurePath._from_parts` to implement this
        method.
        """
        drv, root, parts = self._parse_args(args) # type: ignore
        self._drv = drv
        self._root = root
        self._parts = parts



"`#include <sys/timerfd.h>`"
from __future__ import annotations
from rsyscall._raw import lib # type: ignore
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS
import enum
import rsyscall.near.types as near
import typing as t
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask
from rsyscall.handle.pointer import Pointer, WrittenPointer

# re-exported
from rsyscall.time import Timespec, Itimerspec

__all__ = [
    "CLOCK",
    "TFD",
    "TFD_TIMER",
    "TimerFileDescriptor",
    "Timespec",
    "Itimerspec",
]

class CLOCK(enum.IntEnum):
    REALTIME = lib.CLOCK_REALTIME
    MONOTONIC = lib.CLOCK_MONOTONIC
    BOOTTIME = lib.CLOCK_BOOTTIME
    REALTIME_ALARM = lib.CLOCK_REALTIME_ALARM
    BOOTTIME_ALARM = lib.CLOCK_BOOTTIME_ALARM

class TFD(enum.IntFlag):
    NONE = 0
    CLOEXEC = lib.EFD_CLOEXEC
    NONBLOCK = lib.EFD_NONBLOCK

class TFD_TIMER(enum.IntFlag):
    NONE = 0
    ABSTIME = lib.TFD_TIMER_ABSTIME
    CANCEL_ON_SET = lib.TFD_TIMER_CANCEL_ON_SET

async def _timerfd_create(sysif: SyscallInterface, clockid: CLOCK, flags: TFD) -> near.FileDescriptor:
    return near.FileDescriptor(await sysif.syscall(SYS.timerfd_create, clockid, flags))

async def _timerfd_settime(sysif: SyscallInterface, fd: near.FileDescriptor,
                           flags: TFD_TIMER,
                           new_value: near.Address, old_value: t.Optional[near.Address]) -> None:
    if old_value is None:
        old_value = 0 # type: ignore
    await sysif.syscall(SYS.timerfd_settime, fd, flags, new_value, old_value)

async def _timerfd_gettime(sysif: SyscallInterface, fd: near.FileDescriptor,
                           curr_value: near.Address) -> None:
    await sysif.syscall(SYS.timerfd_gettime, fd, curr_value)


T_fd = t.TypeVar('T_fd', bound='TimerFileDescriptor')
class TimerFileDescriptor(BaseFileDescriptor):
    @t.overload
    async def timerfd_settime(
            self, flags: TFD_TIMER, new_value: WrittenPointer[Itimerspec]) -> None: ...

    @t.overload
    async def timerfd_settime(
            self, flags: TFD_TIMER, new_value: WrittenPointer[Itimerspec],
            old_value: Pointer[Itimerspec]) -> Pointer[Itimerspec]: ...

    async def timerfd_settime(
            self, flags: TFD_TIMER, new_value: WrittenPointer[Itimerspec],
            old_value: t.Optional[Pointer[Itimerspec]]=None) -> t.Optional[Pointer[Itimerspec]]:
        self._validate()
        with new_value.borrow(self.task):
            if old_value:
                with old_value.borrow(self.task):
                    await _timerfd_settime(
                        self.task.sysif, self.near, flags, new_value.near, old_value.near)
                    return old_value
            else:
                await _timerfd_settime(
                    self.task.sysif, self.near, flags, new_value.near, None)
                return None

    async def timerfd_gettime(self, curr_value: Pointer[Itimerspec]) -> Pointer[Itimerspec]:
        self._validate()
        with curr_value.borrow(self.task):
            await _timerfd_gettime(self.task.sysif, self.near, curr_value.near)
            return curr_value

class TimerfdTask(FileDescriptorTask[T_fd]):
    async def timerfd_create(self, clockid: CLOCK, flags: TFD=TFD.NONE) -> T_fd:
        return self.make_fd_handle(await _timerfd_create(self.sysif, clockid, flags|TFD.CLOEXEC))



"`#include <sys/socket.h>`"
from __future__ import annotations
from dataclasses import dataclass
from rsyscall._raw import lib, ffi # type: ignore
from rsyscall.handle.fd import BaseFileDescriptor
from rsyscall.handle.pointer import Pointer
from rsyscall.near.sysif import SyscallInterface
from rsyscall.struct import Struct, Serializable
from rsyscall.sys.syscall import SYS
from rsyscall.time import Timespec
import enum
import rsyscall.near.types as near
import typing as t

class S_IF(enum.IntEnum):
    SOCK = lib.S_IFSOCK
    LNK = lib.S_IFLNK
    REG = lib.S_IFREG
    BLK = lib.S_IFBLK
    DIR = lib.S_IFDIR
    CHR = lib.S_IFCHR
    IFO = lib.S_IFIFO
    # mask with all set
    MT = lib.S_IFMT

class Mode(int):
    def __repr__(self) -> str:
        return oct(self)

    def __str__(self) -> str:
        return repr(self)

@dataclass
class TypeMode:
    # should change this to DT I guess?
    # maybe have an as_dt?
    type: S_IF
    mode: Mode

    @staticmethod
    def from_int(val: int) -> TypeMode:
        type_bits = val & S_IF.MT
        return TypeMode(S_IF(type_bits), Mode(val ^ type_bits))

    def __int__(self) -> int:
        return self.type | self.mode

@dataclass
class Stat(Struct):
    dev: int
    ino: int
    mode: TypeMode
    nlink: int
    uid: int
    gid: int
    rdev: int
    size: int
    blksize: int
    blocks: int
    atime: Timespec
    mtime: Timespec
    ctime: Timespec

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct stat const*', {
            "st_dev": self.dev,
            "st_ino": self.ino,
            "st_nlink": self.nlink,
            "st_mode": self.mode,
            "st_uid": self.uid,
            "st_gid": self.gid,
            "st_rdev": self.rdev,
            "st_size": self.size,
            "st_blksize": self.blksize,
            "st_blocks": self.blocks,
            "st_atim": self.atime._to_cffi_dict(),
            "st_mtim": self.mtime._to_cffi_dict(),
            "st_ctim": self.ctime._to_cffi_dict(),
        })))

    @property
    def atim(self) -> Timespec:
        return self.atime

    @property
    def mtim(self) -> Timespec:
        return self.mtime

    @property
    def ctim(self) -> Timespec:
        return self.ctime

    T = t.TypeVar('T', bound='Stat')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct stat*', ffi.from_buffer(data))
        return cls(
            dev=struct.st_dev,
            ino=struct.st_ino,
            mode=struct.st_mode,
            nlink=struct.st_nlink,
            uid=struct.st_uid,
            gid=struct.st_gid,
            rdev=struct.st_rdev,
            size=struct.st_size,
            blksize=struct.st_blksize,
            blocks=struct.st_blocks,
            atime=Timespec.from_cffi(struct.st_atim),
            mtime=Timespec.from_cffi(struct.st_mtim),
            ctime=Timespec.from_cffi(struct.st_ctim),
        )

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct stat')

async def _fstat(sysif: SyscallInterface, fd: near.FileDescriptor, statbuf: near.Address) -> None:
    await sysif.syscall(SYS.fstat, fd, statbuf)

class StatFileDescriptor(BaseFileDescriptor):
    async def fstat(self, statbuf: Pointer[Stat]) -> Pointer[Stat]:
        self._validate()
        with statbuf.borrow(self.task):
            await _fstat(self.task.sysif, self.near, statbuf.near)
        return statbuf
    

#### Tests ####
from unittest import TestCase
class TestStat(TestCase):
    def test_stat(self) -> None:
        initial = Stat(
            dev=0,
            ino=0,
            mode=TypeMode(S_IF.REG, Mode(0o777)),
            nlink=0,
            uid=0,
            gid=0,
            rdev=0,
            size=0,
            blksize=0,
            blocks=0,
            atime=Timespec(sec=0, nsec=0),
            mtime=Timespec(sec=0, nsec=0),
            ctime=Timespec(sec=0, nsec=0),
        )
        output = Stat.from_bytes(initial.to_bytes())
        self.assertEqual(initial, output)



"`#include <linux/rtnetlink.h>`"
import pyroute2.netlink.rtnl as rtnl
import enum

__all__ = [
    "RTMGRP",
]

class RTMGRP(enum.IntFlag):
    NONE = rtnl.RTMGRP_NONE
    LINK = rtnl.RTMGRP_LINK
    NOTIFY = rtnl.RTMGRP_NOTIFY
    NEIGH = rtnl.RTMGRP_NEIGH
    TC = rtnl.RTMGRP_TC
    IPV4_IFADDR = rtnl.RTMGRP_IPV4_IFADDR
    IPV4_MROUTE = rtnl.RTMGRP_IPV4_MROUTE
    IPV4_ROUTE = rtnl.RTMGRP_IPV4_ROUTE
    IPV4_RULE = rtnl.RTMGRP_IPV4_RULE
    IPV6_IFADDR = rtnl.RTMGRP_IPV6_IFADDR
    IPV6_MROUTE = rtnl.RTMGRP_IPV6_MROUTE
    IPV6_ROUTE = rtnl.RTMGRP_IPV6_ROUTE
    IPV6_IFINFO = rtnl.RTMGRP_IPV6_IFINFO
    DECnet_IFADDR = rtnl.RTMGRP_DECnet_IFADDR
    NOP2 = rtnl.RTMGRP_NOP2
    DECnet_ROUTE = rtnl.RTMGRP_DECnet_ROUTE
    DECnet_RULE = rtnl.RTMGRP_DECnet_RULE
    NOP4 = rtnl.RTMGRP_NOP4
    IPV6_PREFIX = rtnl.RTMGRP_IPV6_PREFIX
    IPV6_RULE = rtnl.RTMGRP_IPV6_RULE
    MPLS_ROUTE = rtnl.RTMGRP_MPLS_ROUTE




"`#include <sys/uio.h>`"
from __future__ import annotations
from rsyscall._raw import lib, ffi # type: ignore
import enum
import typing as t
from rsyscall.handle.pointer import Pointer, WrittenPointer
from rsyscall.struct import Serializable

class RWF(enum.IntFlag):
    NONE = 0
    DSYNC = lib.RWF_DSYNC
    HIPRI = lib.RWF_HIPRI
    SYNC = lib.RWF_SYNC

class IovecList(t.List[Pointer], Serializable):
    def split(self, n: int) -> t.Tuple[IovecList, IovecList]:
        first, middle, second = self.split_with_middle(n)
        if middle is None:
            return first, second
        else:
            first_mid, second_mid = middle
            return IovecList(first + [first_mid]), IovecList([second_mid] + second)

    def split_with_middle(self, n: int) -> t.Tuple[IovecList, t.Optional[t.Tuple[Pointer, Pointer]], IovecList]:
        valid: t.List[Pointer] = []
        middle: t.Optional[t.Tuple[Pointer, Pointer]] = None
        invalid: t.List[Pointer] = []
        for ptr in self:
            size = ptr.size()
            if n >= size:
                valid.append(ptr)
                n -= size
            elif n > 0:
                middle = ptr.split(n)
                n = 0
            else:
                invalid.append(ptr)
        return IovecList(valid), middle, IovecList(invalid)

    def to_bytes(self) -> bytes:
        ret = b""
        for ptr in self:
            ret += bytes(ffi.buffer(ffi.new('struct iovec const*', {
                "iov_base": ffi.cast('void*', int(ptr.near)),
                "iov_len": ptr.size(),
            })))
        return ret

    T = t.TypeVar('T', bound='IovecList')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        raise Exception("can't get pointer handles from raw bytes")

def split_iovec(iov: WrittenPointer[IovecList], ret: int
) -> t.Tuple[WrittenPointer[IovecList], t.Optional[t.Tuple[Pointer, Pointer]], WrittenPointer[IovecList]]:
    first, middle, last = iov.value.split_with_middle(ret)
    if middle is None:
        first_count = len(first)
    else:
        # we include the partially-consumed middle pointer in the first WP[IovecList] returned;
        # I think this is the most ergonomic choice, since in the normal case, I'll want to
        # operate on middle separately, then on second.
        first_count = len(first) + 1
    # TODO this is fairly ad-hoc, splitting on a WrittenPointer should really call into the
    # Serializer to determine validity. (And then I suppose we would have .degrade to degrade a
    # WrittenPointer back into a Pointer so we can split it??? Hmm, seems awkward...)
    first_ptr, last_ptr = iov.split(first_count * ffi.sizeof('struct iovec'))
    return first_ptr._wrote(first), middle, last_ptr._wrote(last)

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor
import contextlib

class UioFileDescriptor(BaseFileDescriptor):
    async def readv(self, iov: WrittenPointer[IovecList], flags: RWF=RWF.NONE
    ) -> t.Tuple[WrittenPointer[IovecList], t.Optional[t.Tuple[Pointer, Pointer]], WrittenPointer[IovecList]]:
        # TODO should check that the WrittenPointer's value and size correspond...
        # maybe we should check that at construction time?
        # otherwise one could make a WrittenPointer that is short, but has a long iovec, and we'd read off the end.
        with contextlib.ExitStack() as stack:
            stack.enter_context(iov.borrow(self.task))
            ret = await _preadv2(self.task.sysif, self.near, iov.near, len(iov.value), -1, flags)
        return split_iovec(iov, ret)

    async def writev(self, iov: WrittenPointer[IovecList], flags: RWF=RWF.NONE
    ) -> t.Tuple[WrittenPointer[IovecList], t.Optional[t.Tuple[Pointer, Pointer]], WrittenPointer[IovecList]]:
        with contextlib.ExitStack() as stack:
            stack.enter_context(iov.borrow(self.task))
            ret = await _pwritev2(self.task.sysif, self.near, iov.near, len(iov.value), -1, flags)
        return split_iovec(iov, ret)

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _preadv2(sysif: SyscallInterface, fd: near.FileDescriptor,
                   iov: near.Address, iovcnt: int, offset: int, flags: RWF) -> int:
    return (await sysif.syscall(SYS.preadv2, fd, iov, iovcnt, offset, flags))

async def _pwritev2(sysif: SyscallInterface, fd: near.FileDescriptor,
                    iov: near.Address, iovcnt: int, offset: int, flags: RWF) -> int:
    return (await sysif.syscall(SYS.pwritev2, fd, iov, iovcnt, offset, flags))



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread

from rsyscall.sys.socket import MSG
from rsyscall.sys.uio import IovecList
from rsyscall.unistd import Pipe
from rsyscall.fcntl import O

class TestPipe(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.pipe = await (await self.thr.task.pipe(await self.thr.malloc(Pipe))).read()

    async def asyncTearDown(self) -> None:
        await self.pipe.read.close()
        await self.pipe.write.close()

    async def test_read_write(self):
        in_data = await self.thr.ptr(b"hello")
        written, _ = await self.pipe.write.write(in_data)
        valid, _ = await self.pipe.read.read(written)
        self.assertEqual(in_data.value, await valid.read())

    async def test_readv_writev(self):
        in_data = [b"hello", b"world"]
        iov = await self.thr.ptr(IovecList([await self.thr.ptr(data) for data in in_data]))
        written, partial, rest = await self.pipe.write.writev(iov)
        read, partial, rest = await self.pipe.read.readv(written)
        self.assertEqual(in_data, [await ptr.read() for ptr in read.value])

    async def test_recv(self) -> None:
        """Sadly, recv doesn't work on pipes

        Which is a major bummer, because that would allow us to avoid
        messing with O_NONBLOCK stuff

        """
        in_data = await self.thr.ptr(b"hello")
        written, _ = await self.pipe.write.write(in_data)
        with self.assertRaises(OSError):
            valid, _ = await self.pipe.read.recv(written, MSG.NONE)



"`#include <unistd.h>`"
from __future__ import annotations
from rsyscall._raw import ffi, lib # type: ignore
import enum
from rsyscall.struct import Serializer, FixedSerializer, Serializable
import struct
import typing as t
import rsyscall.near.types as near
import os
from rsyscall.handle.pointer import Pointer, WrittenPointer, ReadablePointer
if t.TYPE_CHECKING:
    from rsyscall.handle import Task, FileDescriptor

__all__ = [
    "SEEK",
    "OK",
    "ArgList",
    "Pipe",
    "FSFileDescriptor",
]

# re-exported
from rsyscall.unistd.io import SEEK
from rsyscall.unistd.pipe import Pipe

class OK(enum.IntFlag):
    "The mode argument to access, faccessat."
    R = lib.R_OK
    W = lib.W_OK
    X = lib.X_OK
    F = lib.F_OK

class RENAME(enum.IntFlag):
    "The flags argument to renameat2"
    NONE = 0
    EXCHANGE = lib.RENAME_EXCHANGE
    NOREPLACE = lib.RENAME_NOREPLACE
    WHITEOUT = lib.RENAME_WHITEOUT

T_arglist = t.TypeVar('T_arglist', bound='ArgList')
class ArgList(t.List[WrittenPointer[t.Union[str, os.PathLike]]], FixedSerializer):
    "A null-terminated list of null-terminated strings, as passed to execve."
    @classmethod
    def get_serializer(cls, task: Task) -> Serializer[T_arglist]:
        return ArgListSerializer()

import struct
class ArgListSerializer(Serializer[T_arglist]):
    def to_bytes(self, arglist: T_arglist) -> bytes:
        ret = b""
        for ptr in arglist:
            ret += struct.Struct("Q").pack(int(ptr.near))
        ret += struct.Struct("Q").pack(0)
        return ret

    def from_bytes(self, data: bytes) -> T_arglist:
        raise Exception("can't get pointer handles from raw bytes")

def _get_near(fd: t.Optional[BaseFileDescriptor]) -> t.Optional[near.FileDescriptor]:
    if fd is None:
        return None
    else:
        fd._validate()
        return fd.near

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask

from rsyscall.fcntl import AT, O
T_fd = t.TypeVar('T_fd', bound='FSFileDescriptor')
class FSFileDescriptor(BaseFileDescriptor):
    async def readlinkat(self, path: WrittenPointer[t.Union[str, os.PathLike]],
                         buf: Pointer) -> t.Tuple[ReadablePointer, Pointer]:
        """read value of a symbolic link

        manpage: readlinkat(2)
        """
        self._validate()
        with path.borrow(self.task):
            with buf.borrow(self.task):
                ret = await _readlinkat(self.task.sysif, self.near, path.near, buf.near, buf.size())
                return buf.readable_split(ret)

    async def faccessat(self, ptr: WrittenPointer[t.Union[str, os.PathLike]], mode: OK, flags: AT=AT.NONE) -> None:
        """check user's permissions for a file

        manpage: faccessat(2)
        """
        self._validate()
        with ptr.borrow(self.task):
            await _faccessat(self.task.sysif, self.near, ptr.near, mode, flags)

    async def openat(self: T_fd, path: WrittenPointer[t.Union[str, os.PathLike]], flags: O, mode=0o644) -> T_fd:
        """open and possibly create a file

        manpage: openat(2)
        """
        self._validate()
        with path.borrow(self.task) as path_n:
            try:
                fd = await _openat(self.task.sysif, self.near, path_n, flags|O.CLOEXEC, mode)
            except OSError as exn:
                exn.filename = path.value
                raise
            return self.task.make_fd_handle(fd)

    async def mkdirat(self, path: WrittenPointer[t.Union[str, os.PathLike]], mode=0o755) -> None:
        """create a directory

        manpage: mkdirat(2)
        """
        self._validate()
        with path.borrow(self.task) as path_n:
            try:
                await _mkdirat(self.task.sysif, self.near, path_n, mode)
            except OSError as exn:
                exn.filename = path.value
                raise

    async def unlinkat(self, path: WrittenPointer[t.Union[str, os.PathLike]], flags: AT=AT.NONE) -> None:
        """delete a name and possibly the file it refers to

        manpage: unlinkat(2)
        """
        with path.borrow(self.task) as path_n:
            try:
                await _unlinkat(self.task.sysif, self.near, path_n, flags)
            except OSError as exn:
                exn.filename = path.value
                raise

    async def rmdirat(self, path: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        """delete a directory

        manpage: unlinkat(2)
        """
        await self.unlinkat(path, AT.REMOVEDIR)

    async def linkat(self, oldpath: WrittenPointer[t.Union[str, os.PathLike]],
                     newdirfd: t.Optional[FSFileDescriptor],
                     newpath: WrittenPointer[t.Union[str, os.PathLike]],
                     flags: AT=AT.NONE) -> None:
        """make a new name for a file

        manpage: linkat(2)
        """
        self._validate()
        with oldpath.borrow(self.task) as oldpath_n:
            with newpath.borrow(self.task) as newpath_n:
                try:
                    await _linkat(self.task.sysif, self.near, oldpath_n, _get_near(newdirfd), newpath_n, flags)
                except OSError as exn:
                    exn.filename = oldpath.value
                    exn.filename2 = (newdirfd, newpath.value) if newdirfd else newpath.value
                    raise

    async def renameat(self, oldpath: WrittenPointer[t.Union[str, os.PathLike]],
                       newdirfd: t.Optional[FSFileDescriptor],
                       newpath: WrittenPointer[t.Union[str, os.PathLike]],
                       flags: RENAME=RENAME.NONE) -> None:
        """change the name or location of a file

        manpage: renameat2(2)
        """
        self._validate()
        with oldpath.borrow(self.task) as oldpath_n:
            with newpath.borrow(self.task) as newpath_n:
                try:
                    await _renameat2(self.task.sysif, self.near, oldpath_n, _get_near(newdirfd), newpath_n, flags)
                except OSError as exn:
                    exn.filename = oldpath.value
                    exn.filename2 = (newdirfd, newpath.value) if newdirfd else newpath.value
                    raise

    async def symlinkat(self, target: WrittenPointer[t.Union[str, os.PathLike]],
                        linkpath: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        """make a new name for a file

        Note that `self` controls where the link is created, not the target of the link; `self` is
        the `newdirfd` argument to symlinkat.

        manpage: symlinkat(2)

        """
        self._validate()
        with target.borrow(self.task) as target_n:
            with linkpath.borrow(self.task) as linkpath_n:
                try:
                    await _symlinkat(self.task.sysif, target_n, self.near, linkpath_n)
                except OSError as exn:
                    exn.filename = target.value
                    exn.filename2 = (self, linkpath.value)
                    raise

    async def fchmod(self, mode: int) -> None:
        """change permissions of a file

        manpage: fchmod(2)
        """
        self._validate()
        await _fchmod(self.task.sysif, self.near, mode)

    async def ftruncate(self, length: int) -> None:
        """truncate a file to a specified length

        manpage: ftruncate(2)
        """
        self._validate()
        await _ftruncate(self.task.sysif, self.near, length)

    # oldfd has to be a valid file descriptor. newfd is not, technically, required to be
    # open, but that's the best practice for avoiding races, so we require it anyway here.
    async def dup3(self, newfd: T_fd, flags: int) -> T_fd:
        self._validate()
        if not newfd.is_only_handle():
            raise Exception("can't dup over newfd", newfd, "there are more handles to it than just ours")
        if self.near == newfd.near:
            # dup3 fails if newfd == oldfd. I guess I'll just work around that.
            return newfd
        await _dup3(self.task.sysif, self.near, newfd.near, flags)
        # newfd is left as a valid pointer to the new file descriptor
        return newfd

    async def dup2(self, newfd: T_fd) -> T_fd:
        """duplicate a file descriptor

        manpage: dup(2)
        """
        return await self.dup3(newfd, 0)

class FSTask(FileDescriptorTask[T_fd]):
    async def readlink(self, path: WrittenPointer[t.Union[str, os.PathLike]],
                       buf: Pointer) -> t.Tuple[ReadablePointer, Pointer]:
        with path.borrow(self) as path_n:
            with buf.borrow(self) as buf_n:
                ret = await _readlinkat(self.sysif, None, path_n, buf_n, buf.size())
                return buf.readable_split(ret)

    async def access(self, path: WrittenPointer[t.Union[str, os.PathLike]], mode: int, flags: int=0) -> None:
        with path.borrow(self) as path_n:
            try:
                await _faccessat(self.sysif, None, path_n, mode, flags)
            except FileNotFoundError as exn:
                exn.filename = path.value
                raise

    async def open(self, path: WrittenPointer[t.Union[str, os.PathLike]], flags: O, mode=0o644) -> T_fd:
        with path.borrow(self) as path_n:
            try:
                fd = await _openat(self.sysif, None, path_n, flags|O.CLOEXEC, mode)
            except FileNotFoundError as exn:
                exn.filename = path.value
                raise
            return self.make_fd_handle(fd)

    async def mkdir(self, path: WrittenPointer[t.Union[str, os.PathLike]], mode=0o755) -> None:
        with path.borrow(self) as path_n:
            await _mkdirat(self.sysif, None, path_n, mode)

    async def unlink(self, path: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        with path.borrow(self) as path_n:
            await _unlinkat(self.sysif, None, path_n, 0)

    async def rmdir(self, path: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        with path.borrow(self) as path_n:
            await _unlinkat(self.sysif, None, path_n, AT.REMOVEDIR)

    async def link(self, oldpath: WrittenPointer[t.Union[str, os.PathLike]],
                   newpath: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        with oldpath.borrow(self) as oldpath_n:
            with newpath.borrow(self) as newpath_n:
                await _linkat(self.sysif, None, oldpath_n, None, newpath_n, 0)

    async def linkat(self,
                     olddirfd: t.Optional[FSFileDescriptor],
                     oldpath: WrittenPointer[t.Union[str, os.PathLike]],
                     newdirfd: t.Optional[FSFileDescriptor],
                     newpath: WrittenPointer[t.Union[str, os.PathLike]],
                     flags: AT=AT.NONE) -> None:
        """make a new name for a file

        See also `FSFileDescriptor.linkat`.

        manpage: linkat(2)
        """
        with oldpath.borrow(self) as oldpath_n:
            with newpath.borrow(self) as newpath_n:
                try:
                    await _linkat(self.sysif, _get_near(olddirfd), oldpath_n, _get_near(newdirfd), newpath_n, flags)
                except OSError as exn:
                    exn.filename = (olddirfd, oldpath.value) if olddirfd else oldpath.value
                    exn.filename2 = (newdirfd, newpath.value) if newdirfd else newpath.value
                    raise

    async def rename(self, oldpath: WrittenPointer[t.Union[str, os.PathLike]],
                     newpath: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        with oldpath.borrow(self) as oldpath_n:
            with newpath.borrow(self) as newpath_n:
                await _renameat2(self.sysif, None, oldpath_n, None, newpath_n, 0)

    async def symlink(self, target: WrittenPointer[t.Union[str, os.PathLike]],
                      linkpath: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        with target.borrow(self) as target_n:
            with linkpath.borrow(self) as linkpath_n:
                await _symlinkat(self.sysif, target_n, None, linkpath_n)

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _readlinkat(sysif: SyscallInterface, dirfd: t.Optional[near.FileDescriptor],
                      path: near.Address, buf: near.Address, bufsiz: int) -> int:
    if dirfd is None:
        dirfd = AT.FDCWD # type: ignore
    return (await sysif.syscall(SYS.readlinkat, dirfd, path, buf, bufsiz))

async def _faccessat(sysif: SyscallInterface, dirfd: t.Optional[near.FileDescriptor],
                     path: near.Address, flags: int, mode: int) -> None:
    if dirfd is None:
        dirfd = AT.FDCWD # type: ignore
    await sysif.syscall(SYS.faccessat, dirfd, path, flags, mode)

async def _openat(sysif: SyscallInterface, dirfd: t.Optional[near.FileDescriptor],
                  path: near.Address, flags: int, mode: int) -> near.FileDescriptor:
    if dirfd is None:
        dirfd = AT.FDCWD # type: ignore
    return near.FileDescriptor(await sysif.syscall(SYS.openat, dirfd, path, flags, mode))

async def _fchmod(sysif: SyscallInterface, fd: near.FileDescriptor, mode: int) -> None:
    await sysif.syscall(SYS.fchmod, fd, mode)

async def _ftruncate(sysif: SyscallInterface, fd: near.FileDescriptor, length: int) -> None:
    await sysif.syscall(SYS.ftruncate, fd, length)

async def _mkdirat(sysif: SyscallInterface,
                   dirfd: t.Optional[near.FileDescriptor], path: near.Address, mode: int) -> None:
    if dirfd is None:
        dirfd = AT.FDCWD # type: ignore
    await sysif.syscall(SYS.mkdirat, dirfd, path, mode)

async def _unlinkat(sysif: SyscallInterface,
                    dirfd: t.Optional[near.FileDescriptor], path: near.Address, flags: int) -> None:
    if dirfd is None:
        dirfd = AT.FDCWD # type: ignore
    await sysif.syscall(SYS.unlinkat, dirfd, path, flags)

async def _linkat(sysif: SyscallInterface,
                  olddirfd: t.Optional[near.FileDescriptor], oldpath: near.Address,
                  newdirfd: t.Optional[near.FileDescriptor], newpath: near.Address,
                  flags: int) -> None:
    if olddirfd is None:
        olddirfd = AT.FDCWD # type: ignore
    if newdirfd is None:
        newdirfd = AT.FDCWD # type: ignore
    await sysif.syscall(SYS.linkat, olddirfd, oldpath, newdirfd, newpath, flags)

async def _renameat2(sysif: SyscallInterface,
                     olddirfd: t.Optional[near.FileDescriptor], oldpath: near.Address,
                     newdirfd: t.Optional[near.FileDescriptor], newpath: near.Address,
                     flags: int) -> None:
    if olddirfd is None:
        olddirfd = AT.FDCWD # type: ignore
    if newdirfd is None:
        newdirfd = AT.FDCWD # type: ignore
    await sysif.syscall(SYS.renameat2, olddirfd, oldpath, newdirfd, newpath, flags)

async def _symlinkat(sysif: SyscallInterface,
                     target: near.Address, newdirfd: t.Optional[near.FileDescriptor], linkpath: near.Address) -> None:
    if newdirfd is None:
        newdirfd = AT.FDCWD # type: ignore
    await sysif.syscall(SYS.symlinkat, target, newdirfd, linkpath)

async def _dup3(sysif: SyscallInterface,
                oldfd: near.FileDescriptor, newfd: near.FileDescriptor, flags: int) -> near.FileDescriptor:
    return near.FileDescriptor(await sysif.syscall(SYS.dup3, oldfd, newfd, flags))



# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/zhbi4fvq3av9h1rc4wd8imzmc2jk0zsv-nix-2.3.10"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/08qdc7hl7gxs1wbk9fynmvrr4zmpaqfl-bzip2-1.0.6.0.2",
    "/nix/store/1pzq6yh1r415ykizk6zw61z5jlkx9994-openssl-1.1.1j",
    "/nix/store/6hhqb94ilrdmh0cx5bhdqb92m6bqkvdj-zlib-1.2.11",
    "/nix/store/85rb8jkj3g9c24k50bjy5vypj45mnq86-nghttp2-1.41.0-lib",
    "/nix/store/w65ydq10abi813fi7d5j7afdcrxj3aqq-bash-4.4-p23",
    "/nix/store/z88sdrj94jhbp7v8vxf1z8xy6vcv8wln-keyutils-1.6.3-lib",
    "/nix/store/f059zv2fl9ml1q53qj5wbwi8xiim28g1-libkrb5-1.18",
    "/nix/store/hhhzkca6bw5d6qh22g2igqx8vfb07a5b-libssh2-1.9.0",
    "/nix/store/kdbkmym34kivwsjkly8d3j5w6rhj2jz9-brotli-1.0.9-lib",
    "/nix/store/1c9xv0m1khm7wwhvg2xa8xkzjwls27gr-curl-7.74.0",
    "/nix/store/2cw85jfh8yj20cj62lcs21y5jfcq28bi-xz-5.2.5",
    "/nix/store/kk5rvl6m9hc3sgmzf7fw5735k6izkbvb-attr-2.4.48",
    "/nix/store/2dx5xnd837631gcdj0z00h7xlv9jp402-acl-2.2.53",
    "/nix/store/ra609mrrl92jirph1zbahcb06pxmjhcl-openssl-1.1.1j-bin",
    "/nix/store/kic4qky3myhflpf8x6zqq332vqj4wa23-openssl-1.1.1j-dev",
    "/nix/store/6kv4p9gnq517qdxhz3cd56fa4waxwgfp-s2n-tls-1.0.0",
    "/nix/store/8pwcw8vwd194sz91iy0lpiv6wizb2xh4-aws-c-common-0.5.2",
    "/nix/store/if0pk349c8c0pflsjad93i93mq489rk7-aws-checksums-0.1.11",
    "/nix/store/wwkm1iy3cwmhjs4a6fnwv2j79iqhc23q-aws-c-cal-0.4.5",
    "/nix/store/smj73q33wc3c3caxvrbmxqlw78i4vdxw-aws-c-io-0.9.1",
    "/nix/store/d37ikmyb8kz18na1vmlqhi85gn93j381-aws-c-event-stream-0.2.7",
    "/nix/store/fmxgxr7kx29aqbzjp662v5mhkrbjvl91-gcc-10.2.0-lib",
    "/nix/store/3r64nnwxn40brxsmvgwrcfj8vdpr65sq-aws-sdk-cpp-1.8.121",
    "/nix/store/89kh3d91b9i9mqh39fj6xdff1sw81m9f-gzip-1.10",
    "/nix/store/9fkdn82nzxgkxyax3m9qdh2knj7apaa5-coreutils-8.32",
    "/nix/store/c3vfwsi6ipdgxn4wlfc5qmfjp4d86cvf-nix-2.3.10-man",
    "/nix/store/dajsdxwm5q94rzfv0fkmjbgcshfhm7if-busybox-1.32.1-x86_64-unknown-linux-musl",
    "/nix/store/dppx4515xywkvfkbm2h9iri83h4y5pnq-libsodium-1.0.18",
    "/nix/store/gp9zgl6sra50wjraisc7hywgm62xmxdq-editline-1.17.0",
    "/nix/store/hxpycgfpdnmnc9iwxakxr037k0bvdq97-gnutar-1.32",
    "/nix/store/pqlxb52qjmg4sq2m80zh27ppqr4l0zhx-libseccomp-2.5.1-lib",
    "/nix/store/sg3h64z6a7mxmap8zgrhhjscpri5b5vn-xz-5.2.5-bin",
    "/nix/store/vw1lcig1bkncmsmhif27h8hq1gaxldhg-bzip2-1.0.6.0.2-bin",
    "/nix/store/wblkmksnrpql1g3vlnc0rd44469kj10k-sqlite-3.34.1",
    "/nix/store/wkl16zs8r7mhvc1khyj7ya2hn30x0hxw-boehm-gc-8.0.4",
    "/nix/store/zhbi4fvq3av9h1rc4wd8imzmc2jk0zsv-nix-2.3.10",
])



"The high-level interfaces to memory"
from __future__ import annotations
from dneio import run_all
from rsyscall.handle import Task, Pointer, WrittenPointer
from rsyscall.memory.transport import MemoryTransport, MemoryGateway
from rsyscall.memory.allocation_interface import AllocationInterface
from rsyscall.memory.allocator import AllocatorInterface
from rsyscall.struct import FixedSize, T_fixed_size, HasSerializer, T_has_serializer, FixedSerializer, T_fixed_serializer, Serializer, PathLikeSerializer, T_pathlike, StrSerializer
from rsyscall.sys.mman import MemoryMapping
import functools
import os
import rsyscall.near.types as near
import rsyscall.far as far

import typing as t

__all__ = [
    "RAM",
    "perform_batch",
]

class BytesSerializer(Serializer[bytes]):
    def to_bytes(self, val: bytes) -> bytes:
        return val

    def from_bytes(self, data: bytes) -> bytes:
        return data

T = t.TypeVar('T')
class RAM:
    """Central user-friendly class for accessing memory.

    Future work: An option to allocate "const" pointers, which we
    could cache and reuse each time they're requested. This would be
    useful for small pieces of memory which are very frequently used.

    """
    def __init__(self, 
                 task: Task,
                 transport: MemoryTransport,
                 allocator: AllocatorInterface,
    ) -> None:
        self.task = task
        self.transport = transport
        self.allocator = allocator

    @t.overload
    async def malloc(self, cls: t.Type[T_fixed_size]) -> Pointer[T_fixed_size]: ...
    @t.overload
    async def malloc(self, cls: t.Type[T_fixed_serializer], size: int) -> Pointer[T_fixed_serializer]: ...
    @t.overload
    async def malloc(self, cls: t.Type[T_pathlike], size: int) -> Pointer[T_pathlike]: ...
    @t.overload
    async def malloc(self, cls: t.Type[str], size: int) -> Pointer[str]: ...
    @t.overload
    async def malloc(self, cls: t.Type[bytes], size: int) -> Pointer[bytes]: ...

    # have to type: ignore because of https://github.com/python/mypy/issues/9420
    async def malloc(self, cls: t.Union[  # type: ignore
            t.Type[T_fixed_size],
            t.Type[T_fixed_serializer],
            t.Type[T_pathlike],
            t.Type[str],
            t.Type[bytes],
    ], size: t.Optional[int]=None,
    ) -> t.Union[
        Pointer[T_fixed_size],
        Pointer[T_fixed_serializer],
        Pointer[T_pathlike],
        Pointer[str],
        Pointer[bytes],
    ]:
        "Allocate a typed space in memory, sized according to the size of the type or an explicit size argument"
        if size is None:
            if not issubclass(cls, FixedSize):
                raise Exception("non-FixedSize cls passed to malloc without specifying size to allocate", cls)
            ptr: Pointer = await self.malloc_serializer(cls.get_serializer(self.task), cls.sizeof(), cls)
            return ptr
        else:
            if issubclass(cls, FixedSize):
                raise Exception("Can't pass a FixedSize cls to malloc and also specify the size argument", cls, size)
            if issubclass(cls, FixedSerializer):
                ptr = await self.malloc_serializer(cls.get_serializer(self.task), size, cls)
                return ptr
            # special-case Path/str/bytes so that they don't have to get wrapped just for rsyscall
            elif issubclass(cls, os.PathLike):
                pathlike_cls = t.cast(t.Type[T_pathlike], cls)
                return await self.malloc_serializer(PathLikeSerializer(pathlike_cls), size, pathlike_cls)
            elif issubclass(cls, str):
                return await self.malloc_serializer(StrSerializer(), size, str)
            elif issubclass(cls, bytes):
                return await self.malloc_serializer(BytesSerializer(), size, bytes)
            else:
                raise Exception("don't know how to find serializer for", cls)


    @t.overload
    async def ptr(self, data: T_has_serializer) -> WrittenPointer[T_has_serializer]: ...
    @t.overload
    async def ptr(self, data: T_pathlike) -> WrittenPointer[T_pathlike]: ...
    @t.overload
    async def ptr(self, data: str) -> WrittenPointer[str]: ...
    @t.overload
    async def ptr(self, data: t.Union[bytes]) -> WrittenPointer[bytes]: ...
    async def ptr(self, data: t.Union[T_has_serializer, T_pathlike, str, bytes],
    ) -> t.Union[
        WrittenPointer[T_has_serializer],
        WrittenPointer[T_pathlike],
        WrittenPointer[str], WrittenPointer[bytes],
    ]:
        "Take some serializable data and return a pointer in memory containing it."
        if isinstance(data, HasSerializer):
            serializer = data.get_self_serializer(self.task)
            data_bytes = serializer.to_bytes(data)
            ptr: Pointer = await self.malloc_serializer(
                serializer, len(data_bytes), type(data))
            return await self._write_to_pointer(ptr, data, data_bytes)
        elif isinstance(data, os.PathLike):
            path_serializer = PathLikeSerializer(type(data))
            data_bytes = path_serializer.to_bytes(data)
            ptr = await self.malloc_serializer(path_serializer, len(data_bytes), type(data))
            return await self._write_to_pointer(ptr, data, data_bytes)
        elif isinstance(data, str):
            str_serializer = StrSerializer()
            data_bytes = str_serializer.to_bytes(data)
            ptr = await self.malloc_serializer(str_serializer, len(data_bytes), type(data))
            return await self._write_to_pointer(ptr, data, data_bytes)
        elif isinstance(data, bytes):
            ptr = await self.malloc(bytes, len(data))
            return await self._write_to_pointer(ptr, data, data)
        else:
            raise Exception("don't know how to serialize data passed to ptr", data)

    async def perform_batch(self, op: t.Callable[[RAM], t.Awaitable[T]],
                                  allocator: AllocatorInterface=None,
    ) -> T:
        """Batches together memory operations performed by a callable.
        
        See the free function of the same name for more documentation.

        """
        if allocator is None:
            allocator = self.allocator
        return await perform_batch(self.task, self.transport, allocator, op)

    async def malloc_serializer(
            self, serializer: Serializer[T], size: int, typ: t.Type[T],
    ) -> Pointer[T]:
        """Allocate a typed space in memory using an explicitly-specified Serializer.

        This is useful only in relatively niche situations.

        """
        mapping, allocation = await self.allocator.malloc(size, alignment=1)
        try:
            return Pointer(mapping, self.transport, serializer, allocation, typ)
        except:
            allocation.free()
            raise

    async def _write_to_pointer(self, ptr: Pointer[T], data: T, data_bytes: bytes) -> WrittenPointer[T]:
        try:
            return await ptr.write(data)
        except:
            ptr.free()
            raise

class NullAllocation(AllocationInterface):
    "An fake allocation for a null pointer."
    def __init__(self, n: int) -> None:
        self.n = n

    def offset(self) -> int:
        return 0

    def size(self) -> int:
        return self.n

    def split(self, size: int) -> t.Tuple[AllocationInterface, AllocationInterface]:
        return NullAllocation(self.size() - size), NullAllocation(size)

    def merge(self, other: AllocationInterface) -> AllocationInterface:
        raise Exception("can't merge")

    def free(self) -> None:
        pass

class LaterAllocator(AllocatorInterface):
    "An allocator which stores allocation requests and returns null pointers."
    def __init__(self) -> None:
        self.allocations: t.List[t.Tuple[int, int]] = []

    async def malloc(self, size: int, alignment: int) -> t.Tuple[MemoryMapping, AllocationInterface]:
        self.allocations.append((size, alignment))
        return (
            MemoryMapping(t.cast(Task, None), near.MemoryMapping(0, size, 4096), far.File()),
            NullAllocation(size),
        )

class NoopTransport(MemoryTransport):
    "A memory transport which doesn't do anything."
    async def read(self, src: Pointer) -> bytes:
        raise Exception("shouldn't try to read")
    async def write(self, dest: Pointer, data: bytes) -> None:
        pass
    def inherit(self, task: Task) -> NoopTransport:
        raise Exception("shouldn't try to inherit")

class PrefilledAllocator(AllocatorInterface):
    "An allocator which has been prefilled with allocations for an exact sequence of calls to malloc."
    def __init__(self, allocations: t.Sequence[t.Tuple[MemoryMapping, AllocationInterface]]) -> None:
        self.allocations = list(allocations)

    async def malloc(self, size: int, alignment: int) -> t.Tuple[MemoryMapping, AllocationInterface]:
        mapping, allocation = self.allocations.pop(0)
        if allocation.size() != size:
            raise Exception("batch operation seems to be non-deterministic, ",
                            "allocating different sizes/in different order on second run")
        return mapping, allocation

class BatchWriteSemantics(RAM):
    "A variant of RAM which stores writes to pointers so they can be performed later."
    def __init__(self, 
                 task: Task,
                 transport: MemoryTransport,
                 allocator: AllocatorInterface,
    ) -> None:
        self.task = task
        self.transport = transport
        self.allocator = allocator
        self.writes: t.List[t.Tuple[Pointer, bytes]] = []

    async def _write_to_pointer(self, ptr: Pointer[T], data: T, data_bytes: bytes) -> WrittenPointer[T]:
        wptr = ptr._wrote(data)
        self.writes.append((wptr, data_bytes))
        return wptr

async def perform_batch(
        task: Task,
        transport: MemoryTransport,
        allocator: AllocatorInterface,
        batch: t.Callable[[RAM], t.Awaitable[T]],
) -> T:
    """Batches together memory operations performed by a callable.

    To use, first define a callable which takes a RAM, and produces
    some result using only the methods of the passed-in RAM. There are
    several requirements on this callable, described below. Pass that
    callable to this function, and we'll magically batch its
    operations together and return what it returns.

    Requirements on the callable:
    - Do not have any side-effects other than calling methods on the
      passed-in RAM.
    - Do not branch based on the numeric values of pointers; indeed,
      better that you just don't branch at all.

    Concretely, we'll call the callable twice. Once with a RAM which
    no-ops all the allocations and writes, so that we can know the
    size of all the requested allocations. Then we'll batch-allocate
    all that memory, and call the callable again with a RAM which just
    returns those batch allocations, and no-ops any writes. After that
    second call completes, we batch-perform all the writes, and return
    the result of that second call from this function.
    
    This is useful, among other cases, when performing a lot of memory
    allocation and writes in anticipation of one or more syscalls
    which will read values from that memory. For example, we can batch
    together writing the argv and envp arguments to execve, along with
    all the pointers referenced by them. This is much more efficient.

    This technique is inspired by tagless final style. Unfortunately
    we can't really implement that in Python, so an interface is all
    we've got.

    We had a more explicit style before, where you explicitly listed
    the sizes of the allocations you wanted, but it was far less
    ergonomic and less robust. This style is nominally less efficient
    in CPU time, but it improves robustness by making it not possible
    to mess up in calculating the size you want to allocate.

    """
    later_allocator = LaterAllocator()
    await batch(RAM(task, NoopTransport(), later_allocator))
    allocations = await allocator.bulk_malloc(later_allocator.allocations)
    sem = BatchWriteSemantics(task, transport, PrefilledAllocator(allocations))
    ret = await batch(sem)

    await run_all([functools.partial(transport.write, dest, data)
                   for dest, data in sem.writes])
    return ret



"`#include <sys/mount.h>`"
from rsyscall._raw import lib # type: ignore
import os
import typing as t
import enum

class MS(enum.IntFlag):
    NONE = 0
    BIND = lib.MS_BIND
    DIRSYNC = lib.MS_DIRSYNC
    LAZYTIME = lib.MS_LAZYTIME
    MANDLOCK = lib.MS_MANDLOCK
    MOVE = lib.MS_MOVE
    NODEV = lib.MS_NODEV
    NOEXEC = lib.MS_NOEXEC
    NOSUID = lib.MS_NOSUID
    RDONLY = lib.MS_RDONLY
    REC = lib.MS_REC
    RELATIME = lib.MS_RELATIME
    REMOUNT = lib.MS_REMOUNT
    SILENT = lib.MS_SILENT
    SLAVE = lib.MS_SLAVE
    STRICTATIME = lib.MS_STRICTATIME
    SYNCHRONOUS = lib.MS_SYNCHRONOUS
    UNBINDABLE = lib.MS_UNBINDABLE

class UMOUNT(enum.IntFlag):
    NONE = 0
    FORCE = lib.MNT_FORCE
    DETACH = lib.MNT_DETACH
    EXPIRE = lib.MNT_EXPIRE
    NOFOLLOW = lib.UMOUNT_NOFOLLOW

#### Classes ####
import rsyscall.far
from rsyscall.handle.pointer import WrittenPointer

class MountTask(rsyscall.far.Task):
    async def mount(self,
                    source: WrittenPointer[t.Union[str, os.PathLike]],
                    target: WrittenPointer[t.Union[str, os.PathLike]],
                    filesystemtype: WrittenPointer[t.Union[str, os.PathLike]],
                    mountflags: MS,
                    data: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        with source.borrow(self):
            with target.borrow(self):
                with filesystemtype.borrow(self):
                    with data.borrow(self):
                        try:
                            return (await _mount(
                                self.sysif,
                                source.near, target.near, filesystemtype.near,
                                mountflags, data.near))
                        except OSError as exn:
                            exn.filename = source.value
                            exn.filename2 = (target.value, filesystemtype.value, data.value)
                            raise

    async def umount(self, target: WrittenPointer[t.Union[str, os.PathLike]], flags: UMOUNT=UMOUNT.NONE) -> None:
        with target.borrow(self):
            await _umount2(self.sysif, target.near, flags)

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _mount(sysif: SyscallInterface, source: near.Address, target: near.Address,
                 filesystemtype: near.Address, mountflags: MS,
                 data: near.Address) -> None:
    await sysif.syscall(SYS.mount, source, target, filesystemtype, mountflags, data)

async def _umount2(sysif: SyscallInterface, target: near.Address, flags: UMOUNT) -> None:
    await sysif.syscall(SYS.umount2, target, flags)



# Generated by the "nixdeps" line in setup.py
import nixdeps
import pathlib
closure = nixdeps.PackageClosure(pathlib.Path("/nix/store/9fkdn82nzxgkxyax3m9qdh2knj7apaa5-coreutils-8.32"), [
    "/nix/store/fnkixi37wfz5nlyzwab2l51p29793a1m-libunistring-0.9.10",
    "/nix/store/ai6j2i00rik53akq8r5pi2nqjvrqi7ky-libidn2-2.3.0",
    "/nix/store/90illc73xfs933d06daq6d41njs8yh66-glibc-2.32-37",
    "/nix/store/kk5rvl6m9hc3sgmzf7fw5735k6izkbvb-attr-2.4.48",
    "/nix/store/2dx5xnd837631gcdj0z00h7xlv9jp402-acl-2.2.53",
    "/nix/store/9fkdn82nzxgkxyax3m9qdh2knj7apaa5-coreutils-8.32",
])



"""Definitions of namespaces and identifiers tagged with a namespace.

In the analogy to near and far pointers, this file is where we define
segment ids and far pointers. A far pointer is a near pointer plus a
segment id. See `rsyscall.near` for more on this analogy.

Note, we don't actually define any far pointers at the moment. We used
to, but we removed them all in favor of handles; handles are more
robust.

"""
from __future__ import annotations
from dataclasses import dataclass
import contextlib
import rsyscall.near
import typing as t
if t.TYPE_CHECKING:
    from rsyscall.handle import Pointer

@dataclass(eq=False)
class File:
    """An opaque representation of a file, possibly referenced by file descriptors or memory mappings.

    Specifically, this represents what in POSIXese is called an "open file description".

    We use this as an identifier, compared by Python object identity ("is"), to track
    whether two fds or mmaps are referring to the same underlying file.

    This doesn't really fit into the analogy to near and far pointers. In that analogy,
    this class would represent the actual memory being addressed. Not sure what to think
    of that.

    """
    
    pass

#### Segment ids
@dataclass(eq=False)
class FDTable:
    """An opaque representation of an existing file descriptor table, compared with "is".

    This is the namespace in which a near.FileDescriptor is valid.

    For debugging purposes, we take `creator_pid` as an argument. But pids don't uniquely
    identify fd tables, because processes can change fd table, such as by calling
    unshare(CLONE.FILES), while leaving other processes behind in their old address space.

    There aren't any useful, efficient identifiers for file descriptor tables, so we
    compare this object with Python object identify ("is") to see whether two file
    descriptor tables referenced by two Tasks are the same.

    """
    creator_pid: int

    def __str__(self) -> str:
        return f"FDTable({self.creator_pid})"

@dataclass(eq=False)
class AddressSpace:
    """An opaque representation of an existing address space, compared with "is".

    This is the namespace in which a near.Address or near.MemoryMapping is valid.

    For debugging purposes, we take `creator_pid` as an argument. But pids don't uniquely
    identify address spaces, because processes can change address space, such as by
    calling execve, while leaving other processes behind in their old address space.

    There aren't any useful, efficient identifiers for address spaces, so we compare this
    object with Python object identify ("is") to see whether two address spaces referenced
    by two Tasks are the same.

    """
    creator_pid: int
    def __str__(self) -> str:
        return f"AddressSpace({self.creator_pid})"

@dataclass(eq=False)
class PidNamespace:
    """An opaque representation of an existing pid namespace, compared with "is"

    This is the namespace in which a near.Process or near.ProcessGroup is valid. Thread
    groups and sessions are also relative to a pid namespace, but we don't currently
    represent those.

    For debugging purposes, we take `creator_pid` as an argument. But pids don't uniquely
    identify pid namespaces, because processes can change pid namespace, such as by
    calling execve, while leaving other processes behind in their old pid namespace.

    There aren't any useful, efficient identifiers for pid namespaces, so we compare this
    object with Python object identify ("is") to see whether two pid namespaces referenced
    by two Tasks are the same.

    """
    creator_pid: int

class NamespaceMismatchError(Exception):
    "An object was used with a task whose namespaces it doesn't match"
    pass

class FDTableMismatchError(NamespaceMismatchError):
    "A file descriptor was used with a Task with a different fd table"
    pass

class AddressSpaceMismatchError(NamespaceMismatchError):
    "A memory address or memory mapping was used with a Task with a different address space"
    pass

#### Segment register
@dataclass(eq=False)
class Task:
    """A wrapper around `SyscallInterface` which tracks the namespaces of the underlying process

    Note that this is a base class for the more fully featured `rsyscall.Task`.

    We store namespace objects to represent the namespaces that we believe that underlying
    processes is in. Since we have complete control over the process, we can make sure
    this belief is accurate, by updating our stored namespaces when the process changes
    namespace. That isn't done here; it's done in handle.Task.

    Currently, we store only one `PidNamespace`. But each process actually has two pid
    namespaces: 

    - the process's own pid namespace, which determines the pids returned from getpid,
      clone, and other syscalls.
    - the pid namespace that new children will be in.

    The two pid namespaces only differ if we call unshare(CLONE.NEWPID). Currently we
    don't do that because unshare(CLONE.NEWPID) makes monitoring children more complex,
    since they can be deleted without leaving a zombie at any time if the pid namespace
    shuts down. But if we did call unshare(CLONE.NEWPID), we'd need to handle this right.

    In the analogy to near and far pointers, this is like a segment register, if a segment
    register was write-only. Then we'd need to maintain the knowledge of what the segment
    register was set to, outside the segment register itself. That's what we do here.

    There actually were systems where segment registers were, if not quite write-only, at
    least expensive to set and expensive to read. For example, x86_64 - the FS and GS
    segment registers can only be set via syscall. If you wanted to use segmentation on
    such systems, you'd probably have a structure much like this one.

    """
    sysif: rsyscall.near.SyscallInterface
    near_process: rsyscall.near.Process
    fd_table: FDTable
    address_space: AddressSpace
    pidns: PidNamespace

    def _borrow_optional(self, stack: contextlib.ExitStack, ptr: t.Optional[Pointer]
    ) -> t.Optional[rsyscall.near.Address]:
        if ptr is None:
            return None
        else:
            stack.enter_context(ptr.borrow(self))
            return ptr.near

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"Task({self.near_process})"

    def __post_init__(self) -> None:
        pass



"""`#include <sys/capability.h>`

See capget(2).

"""
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.struct import Struct, bits
from dataclasses import dataclass
import typing as t
import enum

__all__ = [
    "CAP",
    "CapHeader",
    "CapData",
]

class CAP(enum.IntEnum):
    CHOWN = lib.CAP_CHOWN
    DAC_OVERRIDE = lib.CAP_DAC_OVERRIDE
    DAC_READ_SEARCH = lib.CAP_DAC_READ_SEARCH
    FOWNER = lib.CAP_FOWNER
    FSETID = lib.CAP_FSETID
    KILL = lib.CAP_KILL
    SETGID = lib.CAP_SETGID
    SETUID = lib.CAP_SETUID
    SETPCAP = lib.CAP_SETPCAP
    LINUX_IMMUTABLE = lib.CAP_LINUX_IMMUTABLE
    NET_BIND_SERVICE = lib.CAP_NET_BIND_SERVICE
    NET_BROADCAST = lib.CAP_NET_BROADCAST
    NET_ADMIN = lib.CAP_NET_ADMIN
    NET_RAW = lib.CAP_NET_RAW
    IPC_LOCK = lib.CAP_IPC_LOCK
    IPC_OWNER = lib.CAP_IPC_OWNER
    SYS_MODULE = lib.CAP_SYS_MODULE
    SYS_RAWIO = lib.CAP_SYS_RAWIO
    SYS_CHROOT = lib.CAP_SYS_CHROOT
    SYS_PTRACE = lib.CAP_SYS_PTRACE
    SYS_PACCT = lib.CAP_SYS_PACCT
    SYS_ADMIN = lib.CAP_SYS_ADMIN
    SYS_BOOT = lib.CAP_SYS_BOOT
    SYS_NICE = lib.CAP_SYS_NICE
    SYS_RESOURCE = lib.CAP_SYS_RESOURCE
    SYS_TIME = lib.CAP_SYS_TIME
    SYS_TTY_CONFIG = lib.CAP_SYS_TTY_CONFIG
    MKNOD = lib.CAP_MKNOD
    LEASE = lib.CAP_LEASE
    AUDIT_WRITE = lib.CAP_AUDIT_WRITE
    AUDIT_CONTROL = lib.CAP_AUDIT_CONTROL
    SETFCAP = lib.CAP_SETFCAP
    MAC_OVERRIDE = lib.CAP_MAC_OVERRIDE
    MAC_ADMIN = lib.CAP_MAC_ADMIN
    SYSLOG = lib.CAP_SYSLOG
    WAKE_ALARM = lib.CAP_WAKE_ALARM
    BLOCK_SUSPEND = lib.CAP_BLOCK_SUSPEND
    AUDIT_READ = lib.CAP_AUDIT_READ
    PERFMON = lib.CAP_PERFMON
    BPF = lib.CAP_BPF
    CHECKPOINT_RESTORE = lib.CAP_CHECKPOINT_RESTORE

def to_uint32s(caps: t.Set[CAP]) -> t.Tuple[int, int]:
    ret: t.List[int] = [0, 0]
    for cap in caps:
        list_idx = (cap // 32)
        uint_idx = (cap % 32)
        ret[list_idx] |= 1 << uint_idx
    return tuple(ret) # type: ignore

def from_uint32s(one: int, two: int) -> t.Set[CAP]:
    return {*{CAP(bit) for bit in bits(one, one_indexed=False)},
            *{CAP(bit+32) for bit in bits(two, one_indexed=False)}}

@dataclass
class CapHeader(Struct):
    # if 0, defaults to calling thread; note that this is a task id, not a thread group id.
    pid: int = 0

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct __user_cap_header_struct*', (lib._LINUX_CAPABILITY_VERSION_3, self.pid))
        return bytes(ffi.buffer(struct))

    T = t.TypeVar('T', bound='CapHeader')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct __user_cap_header_struct*', ffi.from_buffer(data))
        return cls(struct.pid)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct __user_cap_header_struct')

@dataclass
class CapData(Struct):
    "struct __user_cap_data_struct, version 3"
    effective: t.Set[CAP]
    permitted: t.Set[CAP]
    inheritable: t.Set[CAP]

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct __user_cap_data_struct[2]')
        capset_names = ['effective', 'permitted', 'inheritable']
        for name in capset_names:
            capset = getattr(self, name)
            one, two = to_uint32s(capset)
            setattr(struct[0], name, one)
            setattr(struct[1], name, two)
        return bytes(ffi.buffer(struct))

    T = t.TypeVar('T', bound='CapData')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct __user_cap_data_struct[2]', ffi.from_buffer(data))
        capset_names = ['effective', 'permitted', 'inheritable']
        capsets: t.List[t.Set[CAP]] = []
        for name in capset_names:
            one, two = getattr(struct[0], name), getattr(struct[1], name)
            capset = from_uint32s(one, two)
            capsets.append(capset)
        return cls(*capsets)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct __user_cap_data_struct[2]')

#### Classes ####
from rsyscall.handle.pointer import Pointer, WrittenPointer, ReadablePointer
import rsyscall.far

class CapabilityTask(rsyscall.far.Task):
    async def capset(self, hdrp: WrittenPointer[CapHeader], datap: WrittenPointer[CapData]) -> None:
        with hdrp.borrow(self):
            with datap.borrow(self):
                await _capset(self.sysif, hdrp.near, datap.near)

    async def capget(self, hdrp: WrittenPointer[CapHeader], datap: Pointer[CapData]) -> ReadablePointer[CapData]:
        with hdrp.borrow(self):
            with datap.borrow(self):
                await _capget(self.sysif, hdrp.near, datap.near)
                return datap._readable()

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _capget(sysif: SyscallInterface, hdrp: near.Address, datap: near.Address) -> None:
    await sysif.syscall(SYS.capget, hdrp, datap)

async def _capset(sysif: SyscallInterface, hdrp: near.Address, datap: near.Address) -> None:
    await sysif.syscall(SYS.capset, hdrp, datap)
